from typing import Literal, Optional

from semver.version import Version

from . import exceptions

MIN_TARGET_VERSION = "2.2.0+rev1"


def __get_variant(ver: Version) -> Optional[Literal["dev", "prod"]]:
    if "dev" in (ver.build or "") or "dev" in (ver.prerelease or ""):
        return "dev"
    if "prod" in (ver.build or "") or "prod" in (ver.prerelease or ""):
        return "prod"
    return None


def get_hup_action_type(device_type: str, current_version: Optional[str], target_version: str):
    """
    getHUPActionType in Python
    ref: https://github.com/balena-io-modules/balena-hup-action-utils/blob/master/lib/index.ts#L67
    """

    try:
        parsed_current_ver = Version.parse(current_version)  # type: ignore
    except Exception:
        raise exceptions.OsUpdateError("Invalid current balenaOS version")

    try:
        parsed_target_ver = Version.parse(target_version)
    except Exception:
        raise exceptions.OsUpdateError("Invalid target balenaOS version")

    if parsed_current_ver.prerelease or parsed_target_ver.prerelease:
        raise exceptions.OsUpdateError("Updates cannot be performed on pre-release balenaOS versions")

    cur_variant = __get_variant(parsed_current_ver)
    target_variant = __get_variant(parsed_target_ver)

    if target_variant is not None and ((cur_variant == "dev") != (target_variant == "dev")):
        raise exceptions.OsUpdateError(
            "Updates cannot be performed between development and production balenaOS variants"
        )

    if Version.parse(target_version).compare(current_version) < 0:
        raise exceptions.OsUpdateError("OS downgrades are not allowed")

    # For 1.x -> 2.x or 2.x to 2.x only
    if parsed_target_ver.major > 1 and Version.parse(target_version).compare(MIN_TARGET_VERSION) < 0:
        raise exceptions.OsUpdateError("Target balenaOS version must be greater than {0}".format(MIN_TARGET_VERSION))

    return "resinhup{from_v}{to_v}".format(from_v=parsed_current_ver.major, to_v=parsed_target_ver.major)



from typing import List, Optional, Union

from .. import exceptions
from ..auth import Auth
from ..balena_auth import request
from ..pine import PineClient
from ..types import AnyObject
from ..types.models import APIKeyInfoType, APIKeyType
from ..utils import merge
from ..settings import Settings
from .application import Application
from .device import Device


class ApiKey:
    """
    This class implements user API key model for balena python SDK.

    """

    def __init__(self, pine: PineClient, settings: Settings):
        self.__application = Application(pine, settings)
        self.__auth = Auth(pine, settings)
        self.__device = Device(pine, settings)
        self.__pine = pine
        self.__settings = settings

    def create(
        self,
        name: str,
        description: Optional[str] = None,
        expiry_date: Optional[str] = None,
    ) -> str:
        """
        This method registers a new api key for the current user with the name given.

        Args:
            name (str): the API key name
            description (Optional[str]): the API key description
            expiry_date (Optional[str]): the API key expiring date

        Returns:
            str: API key

        Examples:
            >>> balena.models.api_key.create_api_key("myApiKey")
            >>> balena.models.api_key.create_api_key("myApiKey", "my api key description")
            >>> balena.models.api_key.create_api_key("myApiKey", "my descr", datetime.datetime.utcnow().isoformat())
        """
        api_key_body = {"name": name}

        if description is not None and isinstance(description, str):
            api_key_body["description"] = description

        if expiry_date is not None and isinstance(expiry_date, str):
            api_key_body["expiryDate"] = expiry_date

        return request(method="POST", path="/api-key/user/full", settings=self.__settings, body=api_key_body).strip('"')

    def get_all(self, options: AnyObject = {}) -> List[APIKeyType]:
        """
        This function gets all API keys.

        Args:
            options (AnyObject): extra pine options to use

        Returns:
            List[APIKeyType]: user API key

        Examples:
            >>> balena.models.api_key.get_all()
        """
        return self.__pine.get(
            {
                "resource": "api_key",
                "options": merge({"$orderby": "name asc"}, options),
            }
        )

    def update(self, id: int, api_key_info: APIKeyInfoType):
        """
        This function updates details of an API key.

        Args:
            id (str): API key id.
            api_key_info (APIKeyInfoType): new API key info.

        Examples:
            >>> balena.models.api_key.update(1296047, {"name":"new name"})
        """

        if api_key_info is None:
            raise exceptions.InvalidParameter("apiKeyInfo", api_key_info)

        if api_key_info.get("name") is not None and api_key_info.get("name") == "":
            raise exceptions.InvalidParameter("apiKeyInfo.name", api_key_info.get("name"))

        body = {
            "description": api_key_info.get("description"),
            "expiry_date": api_key_info.get("expiry_date"),
        }

        name = api_key_info.get("name")
        if name is not None:
            body["name"] = name

        self.__pine.patch({"resource": "api_key", "id": id, "body": body})

    def revoke(self, id: int):
        """
        This function revokes an API key.

        Args:
            id (int): API key id.

        Examples:
            >>> balena.models.api_key.revoke(1296047)
        """

        self.__pine.delete({"resource": "api_key", "id": id})

    def get_provisioning_api_keys_by_application(
        self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[APIKeyType]:
        """
        Get all provisioning API keys for an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Examples:
            >>> balena.models.api_key.get_provisioning_api_keys_by_application(1296047)
            >>> balena.models.api_key.get_provisioning_api_keys_by_application("myorg/myapp")
        """

        app = self.__application.get(slug_or_uuid_or_id, {"$select": "actor"})
        return self.get_all(merge({"$filter": {"is_of__actor": app.get("actor")}}, options))

    def get_device_api_keys_by_device(self, uuid_or_id: Union[str, int], options: AnyObject = {}) -> List[APIKeyType]:
        """
        Get all API keys for a device.

        Args:
            device_uuid (Union[str, int]): device, uuid (string) or id (int)
            options (AnyObject): extra pine options to use

        Examples:
            >>> balena.models.api_key.get_device_api_keys_by_device("44cc9d186")
            >>> balena.models.api_key.get_device_api_keys_by_device(1111386)
        """

        dev = self.__device.get(uuid_or_id, {"$select": "actor"})
        return self.get_all(merge({"$filter": {"is_of__actor": dev["actor"]}}, options))

    def get_all_named_user_api_keys(self, options: AnyObject = {}) -> List[APIKeyType]:
        """
        Get all named user API keys of the current user.

        Args:
            options (AnyObject): extra pine options to use

        Examples:
            >>> balena.models.api_key.get_all_named_user_api_keys()
        """

        return self.get_all(
            merge(
                {
                    "$filter": {
                        "is_of__actor": self.__auth.get_actor_id(),
                        "name": {"$ne": None},
                    }
                },
                options,
            )
        )



import json
from collections import defaultdict
from threading import Thread
from urllib.parse import urljoin
from typing import Union, Optional, Literal, Callable, TypedDict, Any, List, cast

from twisted.internet import reactor
from twisted.internet.protocol import Protocol
from twisted.web.client import Agent
from twisted.web.http_headers import Headers

from .settings import Settings
from .models.device import Device
from .balena_auth import request, get_token
from .pine import PineClient


class Log(TypedDict):
    message: str
    createdAt: int
    timestamp: int
    isStdErr: bool
    isSystem: bool
    serviceId: Optional[int]


class StreamingParser(Protocol):
    """
    This is low level class and is not meant to be used by end users directly.
    """

    def __init__(self, callback, error):
        self.callback = callback
        self.error = error
        self.pending = b""
        self.is_running = True

    def dataReceived(self, data):
        obj = {}
        self.pending += data

        lines = self.pending.split(b"\n")
        self.pending = lines.pop()

        for line in lines:
            try:
                if line:
                    obj = json.loads(line)
            except Exception as e:
                self.transport.stopProducing()  # type: ignore
                self.transport.loseConnection()  # type: ignore

                if self.error:
                    self.error(e)
                break

            if self.is_running:
                self.callback(obj)

    def connectionLost(self, reason):
        pass


def cbRequest(response, callback, error):
    protocol = StreamingParser(callback, error)
    response.deliverBody(protocol)
    return protocol


def cbDrop(protocol):
    protocol.is_running = False
    protocol.transport.stopProducing()
    protocol.transport.loseConnection()


class Subscription:
    """
    This is low level class and is not meant to be used by end users directly.
    """

    def __init__(self, settings: Settings):
        self.__settings = settings

    def add(
        self,
        uuid: str,
        callback: Callable[[Log], None],
        error: Optional[Callable[[Any], None]] = None,
        count: Optional[Union[int, Literal["all"]]] = None,
    ):
        query = "stream=1"
        if count:
            query = f"stream=1&count={count}"

        url = urljoin(cast(str, self.__settings.get("api_endpoint")), f"/device/v2/{uuid}/logs?{query}")
        headers = Headers({"Authorization": [f"Bearer {get_token(self.__settings)}"]})

        agent = Agent(reactor)
        req = agent.request(b"GET", url.encode(), headers, None)
        req.addCallback(cbRequest, callback, error)
        self.run()

        return req

    def run(self):
        if not reactor.running:  # type: ignore
            Thread(target=reactor.run, args=(False,)).start()  # type: ignore

    def stop(self, d):
        reactor.callFromThread(d.addCallback, cbDrop)  # type: ignore

    def stop_all(self):
        reactor.stop()  # type: ignore


class Logs:
    """
    This class implements functions that allow processing logs from device.

    """

    def __init__(self, pine: PineClient, settings: Settings):
        self.__subscriptions = defaultdict(list)
        self.__settings = settings
        self.__device = Device(pine, settings)
        self.__subscription_handler = Subscription(settings)

    def __exit__(self, exc_type, exc_value, traceback):
        reactor.stop()  # type: ignore

    def subscribe(
        self,
        uuid_or_id: Union[str, int],
        callback: Callable[[Log], None],
        error: Optional[Callable[[Any], None]] = None,
        count: Optional[Union[int, Literal["all"]]] = None,
    ) -> None:
        """
        Subscribe to device logs.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            callback (Callable[[Log], None]): this callback is called on receiving a message.
            error (Optional[Callable[[Any], None]]): this callback is called on an error event.
            count (Optional[Union[int, Literal["all"]]]): number of historical messages to include.
        """

        uuid = self.__device.get(uuid_or_id, {"$select": "uuid"})["uuid"]
        self.__subscriptions[uuid].append(self.__subscription_handler.add(uuid, callback, error, count))

    def history(self, uuid_or_id: Union[str, int], count: Optional[Union[int, Literal["all"]]] = None) -> List[Log]:
        """
        Get device logs history.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            count (Optional[Union[int, Literal["all"]]]): number of historical messages to include.
        """
        uuid = self.__device.get(uuid_or_id, {"$select": "uuid"})["uuid"]
        qs = {}
        if count is not None:
            qs["count"] = count

        return request(method="GET", settings=self.__settings, path=f"/device/v2/{uuid}/logs", qs=qs)

    def unsubscribe(self, uuid_or_id: Union[str, int]) -> None:
        """
        Unsubscribe from device logs for a specific device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
        """
        uuid = self.__device.get(uuid_or_id, {"$select": "uuid"})["uuid"]
        if uuid in self.__subscriptions:
            for d in self.__subscriptions[uuid]:
                self.__subscription_handler.stop(d)
            del self.__subscriptions[uuid]

    def unsubscribe_all(self) -> None:
        """
        Unsubscribe all subscribed devices.
        """
        for device in self.__subscriptions:
            for d in self.__subscriptions[device]:
                self.__subscription_handler.stop(d)
        self.__subscriptions = {}

    def stop(self) -> None:
        """
        Will grecefully unsubscribe from all devices and stop the consumer thread.
        """
        self.unsubscribe_all()
        self.__subscription_handler.stop_all()



import numbers
import re
from collections import defaultdict
from typing import Any, Callable, Dict, Literal, Optional, TypeVar
from .types.models import TypeDevice, TypeDeviceWithServices

from semver.version import Version

from .exceptions import RequestError, SupervisorLocked

SUPERVISOR_LOCKED_STATUS_CODE = 423


def is_id(value: Any) -> bool:
    """
    Return True, if the input value is a valid ID. False otherwise.

    """

    if isinstance(value, numbers.Number):
        try:
            int(value)  # type: ignore
            return True
        except ValueError:
            return False
    return False


def is_full_uuid(value: Any) -> bool:
    """
    Return True, if the input value is a valid UUID. False otherwise.

    """

    if isinstance(value, str):
        if len(value) == 32 or len(value) == 62:
            try:
                str(value)
                return True
            except ValueError:
                return False
    return False


def compare(a, b):
    """

    Return 1 if a is greater than b, 0 if a is equal to b and -1 otherwise.

    """

    if a > b:
        return 1

    if b > a:
        return -1

    return 0


known_pine_option_keys = set(["$select", "$expand", "$filter", "$orderby", "$top", "$skip", "$count"])


def merge(defaults, extras=None, replace_selects=False):
    if extras is None:
        return defaults

    unknown_pine_option = next((key for key in extras if key not in known_pine_option_keys), None)
    if unknown_pine_option is not None:
        raise ValueError(f"Unknown pine option: {unknown_pine_option}")

    result = {**defaults}

    if extras.get("$select"):
        extra_select = (
            extras["$select"]
            if isinstance(extras["$select"], list) or extras["$select"] == "*"
            else [extras["$select"]]
        )
        if replace_selects:
            result["$select"] = extra_select
        elif extra_select == "*":
            result["$select"] = "*"
        else:
            existing_select = result.get("$select")
            existing_select = [existing_select] if not isinstance(existing_select, list) else existing_select
            extra_select = extra_select or []
            merged_select = existing_select + extra_select
            result["$select"] = list(set(merged_select))

    for key in known_pine_option_keys:
        if key in extras:
            result[key] = extras[key]

    if extras.get("$filter"):
        result["$filter"] = (
            {"$and": [defaults.get("$filter", {}), extras["$filter"]]} if defaults.get("$filter") else extras["$filter"]
        )

    if extras.get("$expand"):
        result["$expand"] = merge_expand_options(defaults.get("$expand"), extras["$expand"], replace_selects)

    return result


def merge_expand_options(default_expand=None, extra_expand=None, replace_selects=False):
    if default_expand is None:
        return extra_expand

    default_expand = convert_expand_to_object(default_expand, True)
    extra_expand = convert_expand_to_object(extra_expand)

    for expand_key in extra_expand:
        default_expand[expand_key] = merge(
            default_expand.get(expand_key, {}), extra_expand[expand_key], replace_selects
        )

    return default_expand


def convert_expand_to_object(expand_option, clone_if_needed=False):
    if expand_option is None:
        return {}

    if isinstance(expand_option, str):
        return {expand_option: {}}

    if isinstance(expand_option, list):
        return {k: v for d in expand_option for k, v in (d.items() if isinstance(d, dict) else {d: {}}.items())}

    unknown_pine_option = next((key for key in expand_option if key not in known_pine_option_keys), None)
    if unknown_pine_option is not None:
        raise ValueError(f"Unknown pine expand options: {unknown_pine_option}")

    return {**expand_option} if clone_if_needed else expand_option


def get_current_service_details_pine_expand(
    expand_release: bool,
) -> Dict[str, Any]:
    return {
        "image_install": {
            "$select": ["id", "download_progress", "status", "install_date"],
            "$filter": {
                "status": {
                    "$ne": "deleted",
                },
            },
            "$expand": {
                "image": {
                    "$select": ["id"],
                    "$expand": {
                        "is_a_build_of__service": {
                            "$select": ["id", "service_name"],
                        },
                    },
                },
                **(
                    {
                        "is_provided_by__release": {
                            "$select": ["id", "commit", "raw_version"],
                        },
                    }
                    if expand_release
                    else {}
                ),
            },
        },
    }


def get_single_install_summary(raw_data: Any) -> Any:
    # TODO: Please compare me to node-sdk version
    """
    Builds summary data for an image install or gateway download

    """

    image = raw_data["image"][0]
    service = image["is_a_build_of__service"][0]
    release = None

    if "is_provided_by__release" in raw_data:
        release = raw_data["is_provided_by__release"][0]

    install = {
        "service_name": service["service_name"],
        "image_id": image["id"],
        "service_id": service["id"],
    }

    if release:
        install["commit"] = release["commit"]

    raw_data.pop("is_provided_by__release", None)
    raw_data.pop("image", None)
    install.update(raw_data)

    return install


def generate_current_service_details(raw_device: TypeDevice) -> TypeDeviceWithServices:
    # TODO: Please compare me to node-sdk version
    grouped_services = defaultdict(list)

    for obj in [get_single_install_summary(i) for i in raw_device.get("image_install", [])]:  # type: ignore
        grouped_services[obj.pop("service_name", None)].append(obj)

    raw_device["current_services"] = dict(grouped_services)  # type: ignore
    raw_device["current_gateway_downloads"] = [  # type: ignore
        get_single_install_summary(i) for i in raw_device.get("gateway_download", [])
    ]
    raw_device.pop("image_install", None)  # type: ignore
    raw_device.pop("gateway_download", None)  # type: ignore

    return raw_device  # type: ignore


def is_provisioned(device: Any) -> bool:
    return (
        device.get("supervisor_version") is not None
        and len(device.get("supervisor_version")) > 0
        and device.get("last_connectivity_event") is not None
    )


T = TypeVar("T")


def with_supervisor_locked_error(fn: Callable[[], T]) -> T:
    try:
        return fn()
    except RequestError as e:
        if e.status_code == SUPERVISOR_LOCKED_STATUS_CODE:
            raise SupervisorLocked()
        raise e


def normalize_balena_semver(os_version: str) -> str:
    """
    safeSemver and trimOsText from resin-semver in Python.
    ref: https://github.com/balena-io-modules/resin-semver/blob/master/src/index.js#L5-L24

    """

    # fix major.minor.patch.rev to use rev as build metadata
    version = re.sub(r"(\.[0-9]+)\.rev", r"\1+rev", os_version)
    # fix major.minor.patch.prod to be treat .dev & .prod as build metadata
    version = re.sub(r"([0-9]+\.[0-9]+\.[0-9]+)\.(dev|prod)", r"\1+\2", version)
    # if there are no build metadata, then treat the parenthesized value as one
    version = re.sub(
        r"([0-9]+\.[0-9]+\.[0-9]+(?:[-\.][0-9a-z]+)*) \(([0-9a-z]+)\)",
        r"\1+\2",
        version,
    )
    # if there are build metadata, then treat the parenthesized value as point value
    version = re.sub(
        r"([0-9]+\.[0-9]+\.[0-9]+(?:[-\+\.][0-9a-z]+)*) \(([0-9a-z]+)\)",
        r"\1.\2",
        version,
    )
    # Remove "Resin OS" and "Balena OS" text
    version = re.sub(r"(resin|balena)\s*os\s*", "", version, flags=re.IGNORECASE)
    # remove optional versioning, eg "(prod)", "(dev)"
    version = re.sub(r"\s+\(\w+\)$", "", version)
    # remove "v" prefix
    version = re.sub(r"^v", "", version)
    return version


def ensure_version_compatibility(
    version: str,
    min_version: str,
    version_type: Literal["supervisor", "host OS"],
) -> None:
    version = normalize_balena_semver(version)

    if version and Version.parse(version) < Version.parse(min_version):
        raise ValueError(f"Incompatible {version_type} version: {version} - must be >= {min_version}")


def get_device_os_semver_with_variant(os_version: str, os_variant: Optional[str] = None):
    if not os_version:
        return None

    version_info = Version.parse(normalize_balena_semver(os_version))

    if not version_info:
        return os_version

    tmp = []
    if version_info.prerelease:
        tmp = version_info.prerelease.split(".")
    if version_info.build:
        tmp = tmp + version_info.build.split(".")

    builds = []
    pre_releases = []

    if version_info.build:
        builds = version_info.build.split(".")

    if version_info.prerelease:
        pre_releases = version_info.prerelease.split(".")

    if os_variant and os_variant not in pre_releases and os_variant not in builds:
        builds.append(os_variant)

    return str(
        Version(
            version_info.major,
            version_info.minor,
            version_info.patch,
            version_info.prerelease,
            ".".join(builds),
        )
    )



from urllib.parse import parse_qs
from typing import cast
import jwt

from . import exceptions
from .balena_auth import request
from .settings import Settings

TOKEN_KEY = "token"


class TwoFactorAuth:
    """
    This class implements basic 2FA functionalities for balena python SDK.

    """

    def __init__(self, settings: Settings):
        self.__settings = settings

    def is_enabled(self) -> bool:
        """
        Check if two-factor authentication is enabled.

        Returns:
            bool: True if enabled. Otherwise False.

        Examples:
            >>> balena.twofactor_auth.is_enabled()
        """
        try:
            token = cast(str, self.__settings.get(TOKEN_KEY))
            token_data = jwt.decode(token, algorithms=["HS256"], options={"verify_signature": False})
            return "twoFactorRequired" in token_data
        except jwt.InvalidTokenError:
            raise exceptions.UnsupportedFeature()

    def is_passed(self) -> bool:
        """
        Check if two-factor authentication challenge was passed.
        If the user does not have 2FA enabled, this will be True.

        Returns:
            bool: True if passed. Otherwise False.

        Examples:
            >>> balena.twofactor_auth.is_passed()
        """
        try:
            token = cast(str, self.__settings.get(TOKEN_KEY))
            token_data = jwt.decode(token, algorithms=["HS256"], options={"verify_signature": False})
            if "twoFactorRequired" in token_data:
                return not token_data["twoFactorRequired"]
            return True
        except jwt.InvalidTokenError:
            raise exceptions.UnsupportedFeature()

    def verify(self, code: str) -> str:
        """
        Verifies two factor authentication.
        Note that this method not update the token automatically.
        You should use balena.twofactor_auth.challenge() when possible, as it takes care of that as well.

        Args:
            code (str): two-factor authentication code.

        Returns:
            str: session token.

        Examples:
            >>> balena.twofactor_auth.verify('123456')
        """
        return request(
            method="POST",
            settings=self.__settings,
            path="auth/totp/verify",
            body={"code": code},
        )

    def get_setup_key(self) -> str:
        """
        Retrieves a setup key for enabling two factor authentication.
        This value should be provided to your 2FA app in order to get a token.
        This function only works if you disable two-factor authentication or log in using Auth Token from dashboard.

        Returns:
            str: setup key.

        Examples:
            >>> balena.twofactor_auth.get_setup_key()
        """
        otp_auth_url = request(
            method="GET",
            settings=self.__settings,
            path="auth/totp/setup",
        )
        return parse_qs(otp_auth_url)["secret"][0]

    def enable(self, code: str) -> str:
        """
        Enable two factor authentication.

        Args:
            code (str): two-factor authentication code.

        Returns:
            str: session token.

        Examples:
            >>> balena.twofactor_auth.enable('123456')
        """
        token = self.verify(code)
        self.__settings.set(TOKEN_KEY, token)
        return token

    def challenge(self, code: str) -> None:
        """
        Challenge two-factor authentication.
        If your account has two-factor authentication enabled and logging in using credentials,
        you need to pass two-factor authentication before being allowed to use other functions.

        Args:
            code (str): two-factor authentication code.

        Examples:
            # You need to enable two-factor authentication on dashboard first.
            # Check if two-factor authentication is passed for current session.
            >>> balena.twofactor_auth.is_passed()
            False
            >>> balena.twofactor_auth.challenge('123456')
            # Check again if two-factor authentication is passed for current session.
            >>> balena.twofactor_auth.is_passed()
            True
        """
        token = self.verify(code)
        self.__settings.set(TOKEN_KEY, token)

    def disable(self, password: str) -> str:
        """
        Disable two factor authentication.
        __Note__: Disable will only work when using a token that has 2FA enabled.

        Args:
            password (str): password.

        Returns:
            str: session token.

        Examples:
            >>> balena.twofactor_auth.disable('your_password')
        """
        token = request(
            method="POST",
            settings=self.__settings,
            path="auth/totp/disable",
            body={"password": password},
        )

        self.__settings.set(TOKEN_KEY, token)
        return token



from .. import exceptions
from ..pine import PineClient
from ..types import AnyObject
from ..types.models import ImageType
from ..utils import merge
from ..settings import Settings


class Image:
    """
    This class implements image model for balena python SDK.
    """

    def __init__(self, pine: PineClient, settings: Settings):
        self.__pine = pine
        self.__settings = settings

    def get(self, id: int, options: AnyObject = {}) -> ImageType:
        """
        Get a specific image.

        Args:
            id (int): image id.
            options (AnyObject): extra pine options to use.

        Returns:
            ImageType: image info.
        """
        base_options = {
            "$select": [
                "id",
                "content_hash",
                "dockerfile",
                "project_type",
                "status",
                "error_message",
                "image_size",
                "created_at",
                "push_timestamp",
                "start_timestamp",
                "end_timestamp",
            ]
        }

        image = self.__pine.get({"resource": "image", "id": id, "options": merge(base_options, options, True)})

        if image is None:
            raise exceptions.ImageNotFound(id)

        return image

    def get_logs(self, id: int) -> str:
        """
        Get the build log from an image.

        Args:
            id (str): image id.

        Returns:
            str: build log.
        """
        return self.get(id, {"$select": "build_log"})["build_log"]



from typing import Any, List, Literal, Optional, TypedDict, Union, TypeVar, Dict

__T = TypeVar("__T")


class PineDeferred(TypedDict):
    __id: int


NavigationResource = Union[List[__T], PineDeferred]
ReverseNavigationResource = Union[List[__T], None]
ConceptTypeNavigationResource = Union[List[__T], int]
OptionalNavigationResource = Union[List[__T], PineDeferred, None]


class WebResource(TypedDict):
    filename: str
    href: str
    content_type: str
    content_disposition: str
    size: int


class UserType(TypedDict):
    id: int
    actor: ConceptTypeNavigationResource["ActorType"]
    created_at: str
    username: str

    organization_membership: ReverseNavigationResource["OrganizationMembershipType"]
    user_application_membership: ReverseNavigationResource["ApplicationMembershipType"]
    team_membership: ReverseNavigationResource["TeamMembershipType"]
    has_direct_access_to__application: ReverseNavigationResource["TypeApplication"]


ApplicationMembershipRoles = Literal["developer", "operator", "observer"]
OrganizationMembershipRoles = Literal["administrator", "member"]


class ApplicationMembershipRoleType(TypedDict):
    id: int
    name: ApplicationMembershipRoles


class OrganizationMembershipRoleType(TypedDict):
    id: int
    name: OrganizationMembershipRoles


class TeamApplicationAccessType(TypedDict):
    id: int
    team: NavigationResource["TeamType"]
    # application
    grants_access_to__application: NavigationResource["TypeApplication"]
    application_membership_role: NavigationResource["ApplicationMembershipRoleType"]


class TeamType(TypedDict):
    id: int
    created_at: str
    name: str

    belongs_to__organization: NavigationResource["OrganizationType"]

    # includes__user
    team_membership: ReverseNavigationResource["TeamMembershipType"]
    # grants_access_to__application
    team_application_access: ReverseNavigationResource["TeamApplicationAccessType"]


class TeamMembershipType(TypedDict):
    id: int
    created_at: str

    user: NavigationResource[UserType]
    # team
    is_member_of__team: NavigationResource["TeamType"]


class PublicDeviceType(TypedDict):
    latitude: str
    longitude: str
    belongs_to__application: NavigationResource["TypeApplication"]
    is_of__device_type: NavigationResource["TypeDevice"]
    was_recently_online: bool


class PublicOrganizationType(TypedDict):
    id: int
    name: str
    handle: str


class ActorType(TypedDict):
    id: int

    is_of__user: OptionalNavigationResource[UserType]
    is_of__application: OptionalNavigationResource["TypeApplication"]
    is_of__device: OptionalNavigationResource["TypeDevice"]
    is_of__public_device: OptionalNavigationResource[PublicDeviceType]
    api_key: OptionalNavigationResource["APIKeyType"]


class ApplicationType:
    id: int
    name: str
    slug: str
    description: Optional[str]
    supports_multicontainer: bool
    supports_web_url: bool
    is_legacy: bool
    requires_payment: bool
    needs__os_version_range: Optional[str]
    maximum_device_count: Optional[int]


class APIKeyType(TypedDict):
    id: int
    created_at: str
    name: str
    description: Optional[str]
    expiry_date: Optional[str]
    is_of__actor: NavigationResource[ActorType]


class ApplicationHostedOnApplication:
    application: NavigationResource["TypeApplication"]
    can_use__application_as_host: NavigationResource["TypeApplication"]


class TypeApplication(TypedDict):
    id: int
    created_at: str
    app_name: str
    actor: ConceptTypeNavigationResource[ActorType]
    slug: str
    uuid: str
    is_accessible_by_support_until__date: str
    is_host: bool
    should_track_latest_release: bool
    is_public: bool
    is_of__class: Literal["fleet", "block", "app"]
    is_archived: bool
    is_discoverable: bool
    is_stored_at__repository_url: str
    public_organization: OptionalNavigationResource[PublicOrganizationType]
    application_type: NavigationResource[ApplicationType]
    is_for__device_type: NavigationResource["DeviceTypeType"]
    depends_on__application: OptionalNavigationResource[ApplicationType]
    organization: NavigationResource["OrganizationType"]
    should_be_running__release: OptionalNavigationResource["ReleaseType"]

    application_config_variable: ReverseNavigationResource["EnvironmentVariableBase"]
    application_environment_variable: ReverseNavigationResource["EnvironmentVariableBase"]
    build_environment_variable: ReverseNavigationResource["EnvironmentVariableBase"]
    application_tag: ReverseNavigationResource["BaseTagType"]
    owns__device: ReverseNavigationResource["TypeDevice"]
    owns__public_device: ReverseNavigationResource[PublicDeviceType]
    owns__release: ReverseNavigationResource["ReleaseType"]
    service: ReverseNavigationResource["ServiceType"]
    is_depended_on_by__application: ReverseNavigationResource[ApplicationType]
    is_directly_accessible_by__user: ReverseNavigationResource[UserType]
    user_application_membership: ReverseNavigationResource["ApplicationMembershipType"]
    team_application_access: ReverseNavigationResource[TeamApplicationAccessType]
    can_use__application_as_host: ReverseNavigationResource[ApplicationHostedOnApplication]


class TypeApplicationWithDeviceServiceDetails(TypeApplication):
    owns__device: List["TypeDeviceWithServices"]  # type: ignore


class APIKeyInfoType(TypedDict, total=False):
    name: str
    description: Optional[str]
    expiry_date: Optional[str]


class InviteeType(TypedDict):
    id: int
    created_at: str
    email: str


class ApplicationInviteType(TypedDict):
    id: int
    message: str
    created_at: str
    invitationToken: str
    application_membership_role: NavigationResource[ApplicationMembershipRoleType]
    invitee: NavigationResource[InviteeType]
    is_invited_to__application: NavigationResource[ApplicationType]


class ApplicationMembershipType(TypedDict):
    id: int
    user: NavigationResource[UserType]
    is_member_of__application: NavigationResource[TypeApplication]
    application_membership_role: NavigationResource[ApplicationMembershipRoleType]


class DeviceManufacturerType(TypedDict):
    created_at: str
    modified_at: str
    id: int
    slug: str
    name: str


class DeviceFamilyType(TypedDict):
    created_at: str
    modified_at: str
    id: int
    slug: str
    name: str
    is_manufactured_by__device_manufacturer: OptionalNavigationResource[DeviceManufacturerType]


class CpuArchitectureType(TypedDict):
    id: int
    slug: str
    is_supported_by__device_type: ReverseNavigationResource["CpuArchitectureType"]


class DeviceTypeAliasType(TypedDict):
    id: int
    is_referenced_by__alias: str
    references__device_type: NavigationResource["DeviceTypeType"]


class DeviceTypeType(TypedDict):
    id: int
    slug: str
    name: str
    is_private: bool
    logo: str
    contract: Any
    belongs_to__device_family: OptionalNavigationResource[DeviceFamilyType]
    is_default_for__application: ReverseNavigationResource[TypeApplication]
    is_of__cpu_architecture: NavigationResource[CpuArchitectureType]
    is_accessible_privately_by__organization: ReverseNavigationResource["OrganizationType"]
    describes_device: ReverseNavigationResource["TypeDevice"]
    device_type_alias: ReverseNavigationResource["DeviceTypeAliasType"]


class TypeDeviceState(TypedDict):
    key: str
    name: str


class ServiceInstanceType(TypedDict):
    id: int
    created_at: str
    service_type: str
    ip_address: str
    last_heartbeat: str


class SupervisorReleaseType(TypedDict):
    created_at: str
    id: int
    supervisor_version: str
    image_name: str
    is_public: bool
    note: Optional[str]

    is_for__device_type: NavigationResource[DeviceTypeType]


class ImageInstallType(TypedDict):
    id: int
    download_progress: Optional[float]
    status: str
    install_date: str

    installs__image: NavigationResource["ImageType"]
    device: NavigationResource["TypeDevice"]
    is_provided_by__release: NavigationResource["ReleaseType"]


class TypeDevice(TypedDict):
    id: int
    actor: ConceptTypeNavigationResource[ActorType]
    created_at: str
    modified_at: str
    custom_latitude: str
    custom_longitude: str
    device_name: str
    download_progress: int
    ip_address: str
    public_address: str
    mac_address: str
    is_accessible_by_support_until__date: str
    is_connected_to_vpn: bool
    is_locked_until__date: str
    is_web_accessible: bool
    is_active: bool
    # This is a computed term
    is_frozen: bool
    is_online: bool
    last_connectivity_event: str
    last_vpn_event: str
    latitude: str
    local_id: str
    location: str
    longitude: str
    note: str
    os_variant: str
    os_version: str
    provisioning_progress: int
    provisioning_state: str
    state: TypeDeviceState
    status: str
    status_sort_index: int
    supervisor_version: str
    uuid: str
    vpn_address: str
    api_heartbeat_state: Literal["online", "offline", "timeout", "unknown"]
    memory_usage: int
    memory_total: int
    storage_block_device: str
    storage_usage: int
    storage_total: int
    cpu_usage: int
    cpu_temp: int
    cpu_id: str
    is_undervolted: bool
    # This is a computed term
    overall_status: Any
    # This is a computed term
    overall_progress: int

    is_of__device_type: NavigationResource[DeviceTypeType]
    belongs_to__application: NavigationResource[TypeApplication]
    belongs_to__user: OptionalNavigationResource[UserType]
    is_running__release: OptionalNavigationResource["ReleaseType"]
    should_be_running__release: OptionalNavigationResource["ReleaseType"]
    is_managed_by__service_instance: OptionalNavigationResource[ServiceInstanceType]
    should_be_managed_by__supervisor_release: OptionalNavigationResource[SupervisorReleaseType]
    device_config_variable: ReverseNavigationResource["EnvironmentVariableBase"]
    device_environment_variable: ReverseNavigationResource["EnvironmentVariableBase"]
    device_tag: ReverseNavigationResource["BaseTagType"]
    service_install: ReverseNavigationResource[ServiceInstanceType]
    image_install: ReverseNavigationResource[ImageInstallType]


class TypeCurrentService(TypedDict):
    id: int
    image_id: int
    service_id: int
    download_progress: int
    status: str
    install_date: str


class TypeDeviceWithServices(TypeDevice):
    current_services: Dict[str, List[TypeCurrentService]]


class DeviceMetricsType(TypedDict):
    memory_usage: int
    memory_total: int
    storage_block_device: str
    storage_usage: int
    storage_total: int
    cpu_usage: int
    cpu_temp: int
    cpu_id: str
    is_undervolted: bool


class OrganizationPrivateDeviceTypeAccess(TypedDict):
    id: int
    organization: NavigationResource["OrganizationType"]
    has_private_access_to__device_type: NavigationResource[TypeDevice]


class OrganizationType(TypedDict):
    id: int
    created_at: str
    name: str
    handle: str
    logo_image: WebResource
    has_past_due_invoice_since__date: str
    application: ReverseNavigationResource[TypeApplication]
    organization_membership: ReverseNavigationResource["OrganizationMembershipType"]
    owns__team: ReverseNavigationResource[TeamType]
    organization__has_private_access_to__device_type: ReverseNavigationResource[OrganizationPrivateDeviceTypeAccess]


class OrganizationInviteType(TypedDict):
    id: int
    message: str
    created_at: str
    invitationToken: str
    organization_membership_role: NavigationResource[OrganizationMembershipRoleType]
    invitee: NavigationResource[InviteeType]
    is_invited_to__organization: NavigationResource[OrganizationType]


class OrganizationMembershipType(TypedDict):
    id: int
    created_at: str
    user: NavigationResource[UserType]
    is_member_of__organization: NavigationResource[OrganizationType]
    organization_membership_role: NavigationResource[OrganizationMembershipRoleType]
    effective_seat_role: str
    organization_membership_tag: ReverseNavigationResource["OrganizationMembershipTagType"]


class OrganizationMembershipTagType(TypedDict):
    organization_membership: NavigationResource[OrganizationMembershipType]


class BaseTagType(TypedDict):
    id: int
    tag_key: str
    value: str


class EnvironmentVariableBase(TypedDict):
    id: int
    name: str
    value: str


class ReleaseImageType(TypedDict):
    id: int
    created_at: str
    image: NavigationResource["ImageType"]
    is_part_of__release: NavigationResource["ReleaseType"]


class ImageType(TypedDict):
    id: int
    created_at: str
    build_log: str
    contract: Any
    content_hash: str
    project_type: str
    status: str
    is_stored_at__image_location: str
    start_timestamp: str
    end_timestamp: str
    push_timestamp: str
    image_size: int
    dockerfile: str
    error_message: str
    is_a_build_of__service: NavigationResource["ServiceType"]
    release_image: ReverseNavigationResource[ReleaseImageType]


class ServiceType(TypedDict):
    id: int
    created_at: str
    service_name: str
    application: NavigationResource[TypeApplication]
    is_built_by__image: ReverseNavigationResource[ImageType]
    service_environment_variable: ReverseNavigationResource[EnvironmentVariableBase]
    device_service_environment_variable: ReverseNavigationResource[EnvironmentVariableBase]


ReleaseStatus = Literal["cancelled", "error", "failed", "interrupted", "local", "running", "success", "timeout"]


class ReleaseVersion(TypedDict):
    raw: str
    major: int
    minor: int
    patch: int
    version: str
    build: List[str]
    prerelease: List[Union[str, int]]


class ReleaseType(TypedDict):
    id: int
    created_at: str
    commit: str
    composition: Any
    contract: Any
    status: ReleaseStatus
    source: str
    build_log: str
    is_invalidated: bool
    start_timestamp: str
    update_timestamp: str
    end_timestamp: str
    phase: Literal["next", "current", "sunset", "end-of-life"]
    semver: str
    semver_major: int
    semver_minor: int
    semver_patch: int
    semver_prerelease: str
    semver_build: str
    variant: str
    revision: int
    known_issue_list: str
    # This is a computed term
    raw_version: str
    # This is a computed term
    version: ReleaseVersion
    is_final: bool
    is_finalized_at__date: str
    note: str
    invalidation_reason: str
    is_created_by__user: OptionalNavigationResource[UserType]
    belongs_to__application: NavigationResource[TypeApplication]
    release_image: ReverseNavigationResource[ReleaseImageType]
    should_be_running_on__application: ReverseNavigationResource[TypeApplication]
    is_running_on__device: ReverseNavigationResource[TypeDevice]
    should_be_running_on__device: ReverseNavigationResource[TypeDevice]
    release_tag: ReverseNavigationResource[BaseTagType]


class ImageBasicInfoType(TypedDict):
    id: int
    service_name: str


class BasicUserInfoType(TypedDict):
    id: int
    username: str


class ReleaseWithImageDetailsType(ReleaseType):
    images: List[ImageBasicInfoType]
    user: BasicUserInfoType


class DeviceHistoryType(TypedDict):
    created_at: str
    id: int
    end_timestamp: str
    is_created_by__actor: OptionalNavigationResource[ActorType]
    is_ended_by__actor: OptionalNavigationResource[ActorType]
    tracks__device: NavigationResource[TypeDevice]
    tracks__actor: OptionalNavigationResource[ActorType]
    uuid: Optional[str]
    belongs_to__application: NavigationResource[TypeApplication]
    is_active: bool
    is_running__release: OptionalNavigationResource[ReleaseType]
    should_be_running__release: OptionalNavigationResource[ReleaseType]
    os_version: Optional[str]
    os_variant: Optional[str]
    supervisor_version: Optional[str]
    is_of__device_type: OptionalNavigationResource[DeviceTypeType]
    should_be_managed_by__release: OptionalNavigationResource[ReleaseType]


class SSHKeyType(TypedDict):
    title: str
    public_key: str
    id: int
    created_at: str
    user: NavigationResource[UserType]


class CreditBundleType(TypedDict):
    id: int
    created_at: str
    is_created_by__user: OptionalNavigationResource[UserType]
    original_quantity: float
    total_balance: float
    total_cost: float
    payment_status: Literal["processing", "paid", "failed", "complimentary", "cancelled", "refunded"]
    belongs_to__organization: NavigationResource[OrganizationType]
    is_for__feature: Any
    is_associated_with__invoice_id: Optional[str]
    error_message: Optional[str]



"""
Welcome to the balena python SDK documentation.
This document aims to describe all the functions supported by the SDK, as well as
showing examples of their expected usage.

Install the Balena SDK:

From Pip:
```
pip install balena-sdk
```

From Source (In case, you want to test a development branch):
```
https://github.com/balena-io/balena-sdk-python
```

Getting started:

```python
>>> from balena import Balena
>>> balena = Balena()
>>> credentials = {'username':<your email>, 'password':<your password>}
>>> balena.auth.login(**credentials)
...
```

The Balena object can be configured with a dict of type Settings

```python
balena = Balena({
    "balena_host": "balena-cloud.com",
    "api_version": "v6",
    "device_actions_endpoint_version": "v1",
    "data_directory": "/home/example/.balena",
    "image_cache_time": str(1 * 1000 * 60 * 60 * 24 * 7), # 1 week
    "token_refresh_interval": str(1 * 1000 * 60 * 60),    # 1 hour
    "timeout": str(30 * 1000),                            # request timeout, 30s
    "request_limit": str(300), # the number of requests per request_limit_interval that the SDK should respect, defaults to unlimited.
    "request_limit_interval": str(60), # the timespan that the request_limit should apply to in seconds, defaults to 60s (1 minute).
    "retry_rate_limited_request": False, # awaits and retry once a request is rate limited (429)
})
```

Notice that if you want to change for the staging environment, you could simply do:
balena = Balena({"balena_host": "balena-staging.com"})

However, this will overwrite your balena-cloud settings (stored api keys etc). So we recommend using
a different data_directory for each balena-sdk instance, e.g:

```python
balena_prod = Balena()
balena_staging = Balena({
    "balena_host": "balena-staging.com",
    "data_directory": "/home/balena-staging-sdk/.balena",
})
```

In adition, you can also run balena-python-sdk completely in memory, without writing anything to the file system like:

```python
balena_prod = Balena({"data_directory": False})
balena_staging = Balena({
    "balena_host": "balena-staging.com",
    "data_directory": False
})
```

By default the SDK will throw once a request is Rate limited by the API (with a 429 status code).
A 429 request will contain a header called "retry-after" which informs how long the client should wait before trying a new request.
If you would like the SDK to use this header and wait and automatically retry the request, just do:

```python
balena = Balena({"retry_rate_limited_request": True})
```

If you feel something is missing, not clear or could be improved, [please don't
hesitate to open an issue in GitHub](https://github.com/balena-io/balena-sdk-python/issues), we'll be happy to help.
"""  # noqa: E501

from typing import Optional
from .auth import Auth
from .logs import Logs
from .models import Models
from .pine import PineClient
from .settings import SettingsConfig, Settings

__version__ = "14.4.3"


class Balena:
    """
    This class implements all functions supported by the python SDK.
    Attributes:
            settings (Settings): configuration settings for balena python SDK.
            logs (Logs): logs from devices working on Balena.
            auth (Auth): authentication handling.
            models (Models): all models in balena python SDK.

    """

    def __init__(self, settings: Optional[SettingsConfig] = None):
        self.settings = Settings(settings)
        self.pine = PineClient(self.settings, __version__)
        self.logs = Logs(self.pine, self.settings)
        self.auth = Auth(self.pine, self.settings)
        self.models = Models(self.pine, self.settings)



from typing import Union, List
from .organization import Organization
from ..utils import merge
from ..pine import PineClient
from ..types import AnyObject
from ..types.models import CreditBundleType
from ..settings import Settings


class CreditBundle:
    """
    This class implements credit bundle model for balena python SDK.
    """

    def __init__(self, pine: PineClient, settings: Settings):
        self.__organization = Organization(pine, settings)
        self.__pine = pine

    def __get_org_id(self, organization: Union[str, int]) -> int:
        return self.__organization.get(organization, {"$select": "id"})["id"]

    def get_all_by_org(self, organization: Union[str, int], options: AnyObject = {}) -> List[CreditBundleType]:
        """
        Get all of the credit bundles purchased by the given org

        Args:
            organization (Union[str, int]): handle (string) or id (number) of the target organization.
            options (AnyObject): extra pine options to use

        Returns:
            List[CreditBundleType]: credit bundles.

        Examples:
            >>> balena.models.credit_bundle.get_all_by_org('myorghandle')
        """
        org_id = self.__get_org_id(organization)
        return self.__pine.get(
            {
                "resource": "credit_bundle",
                "options": merge(
                    options, {"$filter": {"belongs_to__organization": org_id}, "$orderby": {"created_at": "desc"}}
                ),
            }
        )

    def create(self, organization: Union[str, int], feature_id: int, credits_to_purchase: float) -> CreditBundleType:
        """
        Purchase a credit bundle for the given feature and org of the given quantity

        Args:
            organization (Union[str, int]): handle (string) or id (number) of the target organization.
            feature_id (int): id (number) of the feature for which credits are being purchased.
            credits_to_purchase (float): number of credits being purchased.

        Returns:
            CreditBundleType: credit bundle.

        Examples:
            >>> balena.models.credit_bundle.create('myorghandle', 1234, 200)
        """
        org_id = self.__get_org_id(organization)
        return self.__pine.post(
            {
                "resource": "credit_bundle",
                "body": {
                    "belongs_to__organization": org_id,
                    "is_for__feature": feature_id,
                    "original_quantity": credits_to_purchase,
                },
            }
        )



from typing import Any, Dict, Literal, TypedDict, Union

AnyObject = Dict[str, Any]
ApplicationMembershipRoles = Literal["developer", "operator", "observer"]


class ShutdownOptions(TypedDict, total=False):
    force: bool


class ApplicationInviteOptions(TypedDict, total=False):
    invitee: str
    roleName: ApplicationMembershipRoles
    message: str


class ResourceKeyDict(TypedDict, total=False):
    user: int
    is_member_of__application: int
    is_member_of__organization: int


ResourceKey = Union[int, ResourceKeyDict]



from typing import Any, Callable, Generic, List, Optional, TypeVar

from .pine import PineClient
from .types import AnyObject
from .utils import is_id, merge

T = TypeVar("T")


class DependentResource(Generic[T]):
    def __init__(
        self,
        resource_name: str,
        resource_key_field: str,
        parent_resource_name: str,
        get_resource_id: Callable[[Any], int],
        pine: PineClient,
    ):
        self.resource_name = resource_name
        self.resource_key_field = resource_key_field
        self.parent_resource_name = parent_resource_name
        self.get_resource_id = get_resource_id
        self.__pine = pine

    def _get_all(self, options: AnyObject = {}) -> List[T]:
        default_orderby = {"$orderby": {self.resource_key_field: "asc"}}

        return self.__pine.get(
            {
                "resource": self.resource_name,
                "options": merge(default_orderby, options),
            }
        )

    def _get_all_by_parent(self, parent_param: Any, options: AnyObject = {}) -> List[T]:
        parent_id = parent_param if is_id(parent_param) else self.get_resource_id(parent_param)

        get_options = {
            "$filter": {self.parent_resource_name: parent_id},
            "$orderby": f"{self.resource_key_field} asc",
        }

        return self._get_all(merge(get_options, options))

    def _get(self, parent_param: Any, key: str) -> Optional[str]:
        parent_id = parent_param if is_id(parent_param) else self.get_resource_id(parent_param)

        dollar_filter = {self.parent_resource_name: parent_id, self.resource_key_field: key}

        result = self.__pine.get(
            {
                "resource": self.resource_name,
                "options": {"$select": "value", "$filter": dollar_filter},
            }
        )

        if len(result) >= 1:
            return result[0].get("value")

    def _set(self, parent_param: Any, key: str, value: str) -> None:
        parent_id = parent_param if is_id(parent_param) else self.get_resource_id(parent_param)

        upsert_id = {self.parent_resource_name: parent_id, self.resource_key_field: key}

        self.__pine.upsert(
            {
                "resource": self.resource_name,
                "id": upsert_id,
                "body": {"value": value},
            }
        )

    def _remove(self, parent_param: Any, key: str) -> None:
        parent_id = parent_param if is_id(parent_param) else self.get_resource_id(parent_param)

        dollar_filter = {self.parent_resource_name: parent_id, self.resource_key_field: key}

        self.__pine.delete({"resource": self.resource_name, "options": {"$filter": dollar_filter}})



import os
from datetime import datetime
from typing import Any, Optional
from urllib.parse import urljoin

import jwt
import requests

from . import exceptions
from .settings import Settings
import balena


def __request_new_token(settings: Settings) -> str:
    headers = {"Authorization": f"Bearer {settings.get('token')}"}
    url = urljoin(settings.get("api_endpoint"), "whoami")
    response = requests.get(url, headers=headers)

    if not response.ok:
        raise exceptions.RequestError(response.content.decode(), response.status_code)

    return response.content.decode()


def __should_update_token(token: str, interval: str) -> bool:
    try:
        # Auth token
        token_data = jwt.decode(token, algorithms=["HS256"], options={"verify_signature": False})
        # dt will be the same as Date.now() in Javascript but converted to
        # milliseconds for consistency with js/sc sdk
        dt = (datetime.utcnow() - datetime.utcfromtimestamp(0)).total_seconds()
        dt = dt * 1000
        age = dt - (int(token_data["iat"]) * 1000)
        return int(age) >= int(interval)
    except jwt.InvalidTokenError:
        return False


def get_token(settings: Settings) -> Optional[str]:
    if settings.has("token"):
        token = settings.get("token")
        interval = settings.get("token_refresh_interval")
        if __should_update_token(token, interval):
            settings.set("token", __request_new_token(settings))
        api_key = settings.get("token")

    else:
        api_key = os.environ.get("BALENA_API_KEY") or os.environ.get("RESIN_API_KEY")

    return api_key


def request(
    method: str,
    path: str,
    settings: Settings,
    body: Optional[Any] = None,
    endpoint: Optional[str] = None,
    token: Optional[str] = None,
    qs: Optional[Any] = {},
    return_raw: bool = False,
    stream: bool = False,
    send_token: bool = True,
) -> Any:
    if endpoint is None:
        endpoint = settings.get("api_endpoint")

    if token is None and send_token:
        token = get_token(settings)

    url = urljoin(endpoint, path)

    if token is None and send_token:
        raise exceptions.NotLoggedIn()
    try:
        headers = {"X-Balena-Client": f"balena-python-sdk/{balena.__version__}"}
        if send_token:
            headers["Authorization"] = f"Bearer {token}"

        req = requests.request(method=method, url=url, params=qs, json=body, headers=headers, stream=stream)

        if return_raw:
            return req

        try:
            return req.json()
        except Exception:
            return req.content.decode()

    except Exception as e:
        if not send_token:
            raise e
        raise exceptions.NotLoggedIn()



from typing import Any, Optional, cast
from urllib.parse import urljoin
from ratelimit import limits, sleep_and_retry
from time import sleep
import mimetypes
import io
import requests
import os
from pine_client import PinejsClientCore
from pine_client.client import Params

from .balena_auth import get_token
from .exceptions import RequestError, InvalidOption
from .settings import Settings


class PineClient(PinejsClientCore):
    def __init__(self, settings: Settings, sdk_version: str, params: Optional[Params] = None):
        if params is None:
            params = {}

        self.__settings = settings
        self.__sdk_version = sdk_version

        api_url = cast(str, settings.get("api_endpoint"))
        api_version = cast(str, settings.get("api_version"))

        try:
            calls = int(self.__settings.get("request_limit"))
            period = int(self.__settings.get("request_limit_interval"))

            self.__request = sleep_and_retry(limits(calls=calls, period=period)(self.__base_request))
        except InvalidOption:
            self.__request = self.__base_request

        super().__init__({**params, "api_prefix": urljoin(api_url, api_version) + "/"})

    def _request(self, method: str, url: str, body: Optional[Any] = None) -> Any:
        return self.__request(method, url, body)

    def __base_request(self, method: str, url: str, body: Optional[Any] = None) -> Any:
        token = get_token(self.__settings)

        headers = {"X-Balena-Client": f"balena-python-sdk/{self.__sdk_version}"}
        if token is not None:
            headers["Authorization"] = f"Bearer {token}"

        is_multipart_form_data = False
        files = {}
        values = {}
        if body is not None:
            for k, v in body.items():
                if isinstance(v, io.BufferedReader):
                    mimetype, _ = mimetypes.guess_type(v.name)
                    files[k] = (os.path.basename(v.name), v, mimetype)
                    is_multipart_form_data = True
                else:
                    values[k] = v

        if is_multipart_form_data:
            req = requests.request(method, url=url, files=files, data=values, headers=headers)
        else:
            req = requests.request(method, url=url, json=body, headers=headers)

        if req.ok:
            try:
                return req.json()
            except Exception:
                return req.content.decode()
        else:
            retry_after = req.headers.get("retry-after")
            if (
                self.__settings.get("retry_rate_limited_request") is True
                and req.status_code == 429
                and retry_after is not None
                and retry_after.isdigit()
            ):
                sleep(int(retry_after))
                return self.__base_request(method, url, body)

            raise RequestError(body=req.content.decode(), status_code=req.status_code)



import binascii
import datetime
import os
from typing import Any, Callable, List, Optional, TypedDict, Union, cast
from urllib.parse import urljoin

from deprecated import deprecated
import requests
from semver.version import Version
from json.decoder import JSONDecodeError

from .. import exceptions
from ..auth import Auth
from ..balena_auth import request
from ..dependent_resource import DependentResource
from ..hup import get_hup_action_type
from ..pine import PineClient
from ..resources import Message
from ..settings import Settings
from ..types import AnyObject
from ..types.models import BaseTagType, DeviceMetricsType, EnvironmentVariableBase, TypeDevice, TypeDeviceWithServices
from ..utils import (
    ensure_version_compatibility,
    generate_current_service_details,
    get_current_service_details_pine_expand,
    get_device_os_semver_with_variant,
    is_full_uuid,
    is_id,
    is_provisioned,
    merge,
    with_supervisor_locked_error,
)
from .application import Application
from .config import Config
from .device_type import DeviceType
from .history import DeviceHistory
from .organization import Organization
from .os import DeviceOs, normalize_balena_semver
from .release import Release

LOCAL_MODE_MIN_OS_VERSION = "2.0.0"
LOCAL_MODE_MIN_SUPERVISOR_VERSION = "4.0.0"
LOCAL_MODE_ENV_VAR = "RESIN_SUPERVISOR_LOCAL_MODE"
OVERRIDE_LOCK_ENV_VAR = "RESIN_OVERRIDE_LOCK"
MIN_SUPERVISOR_MC_API = "7.0.0"
MIN_OS_MC = "2.12.0"
MIN_SUPERVISOR_APPS_API = "1.8.0-alpha.0"


class LocationType(TypedDict):
    latitude: Union[str, int]
    longitude: Union[str, int]


class RegisterResponse(TypedDict):
    id: int
    uuid: str
    api_key: str


class LocalModeResponse(TypedDict):
    message: str
    supported: bool


class HUPStatusResponse(TypedDict):
    status: str
    lastRun: int
    action: str
    parameters: AnyObject


class SupervisorStateType(TypedDict):
    api_port: str
    ip_address: str
    os_version: str
    supervisor_version: str
    update_pending: bool
    update_failed: bool
    update_downloaded: bool
    status: str
    commit: str
    download_progress: str


class DeviceStatus:
    """
    Balena device statuses.
    """

    IDLE = "Idle"
    CONFIGURING = "Configuring"
    UPDATING = "Updating"
    OFFLINE = "Offline"
    POST_PROVISIONING = "Post Provisioning"
    INACTIVE = "Inactive"


class Device:
    """
    This class implements device model for balena python SDK.
    """

    def __init__(self, pine: PineClient, settings: Settings):
        self.__pine = pine
        self.__settings = settings
        self.__config = Config(settings)
        self.__auth = Auth(pine, settings)
        self.__application = Application(pine, settings)
        self.__release = Release(pine, settings)
        self.__device_os = DeviceOs(pine, settings)
        self.__device_type = DeviceType(pine, settings)
        self.__organization = Organization(pine, settings)
        self.__LOCAL_MODE_SELECT = [
            "id",
            "os_version",
            "os_variant",
            "supervisor_version",
            "last_connectivity_event",
        ]

        self.tags = DeviceTag(pine, self, self.__application)
        self.config_var = DeviceConfigVariable(pine, self, self.__application)
        self.env_var = DeviceEnvVariable(pine, self, self.__application)
        self.service_var = DeviceServiceEnvVariable(pine, self, self.__application)
        self.history = DeviceHistory(pine, settings)

        self.__supervisor_address = os.environ.get("BALENA_SUPERVISOR_ADDRESS")
        self.__supervisor_api_key = os.environ.get("BALENA_SUPERVISOR_API_KEY")
        self.__on_device_app_id = os.environ.get("BALENA_APP_ID")
        self.__local_device_uuid = os.environ.get("BALENA_DEVICE_UUID")

        self.__on_device = all([self.__supervisor_address, self.__supervisor_api_key, self.__on_device_app_id])

    def __get_applied_device_config_variable_value(self, uuid_or_id: Union[str, int], name: str):
        options = {
            "$expand": {
                "device_config_variable": {
                    "$select": "value",
                    "$filter": {"name": name},
                },
                "belongs_to__application": {
                    "$select": "id",
                    "$expand": {
                        "application_config_variable": {
                            "$select": "value",
                            "$filter": {"name": name},
                        }
                    },
                },
            }
        }

        result = self.get(uuid_or_id, options)

        device_config = next(iter(result["device_config_variable"]), None)
        app_config = next(
            iter(result["belongs_to__application"][0]["application_config_variable"]),
            None,
        )

        if device_config is not None:
            return device_config.get("value")

        if app_config is not None:
            return app_config.get("value")

        return None

    def __set(
        self,
        uuid_or_id_or_ids: Union[str, int, List[int]],
        body: Any,
        fn: Optional[Callable] = None,
    ) -> None:
        if fn is None:
            fn = self.__pine.patch

        if isinstance(uuid_or_id_or_ids, (int, str)):
            is_potentially_full_uuid = is_full_uuid(uuid_or_id_or_ids)
            if is_potentially_full_uuid or is_id(uuid_or_id_or_ids):
                fn(
                    {
                        "resource": "device",
                        "id": uuid_or_id_or_ids if is_id(uuid_or_id_or_ids) else {"uuid": uuid_or_id_or_ids},
                        "body": body,
                    }
                )
            else:
                fn(
                    {
                        "resource": "device",
                        "options": {"$filter": {"uuid": {"$startswith": uuid_or_id_or_ids}}},
                        "body": body,
                    }
                )

        else:
            chunk_size = 200
            chunked_devices = [
                uuid_or_id_or_ids[i : i + chunk_size]  # noqa: E203 type: ignore
                for i in range(0, len(uuid_or_id_or_ids), chunk_size)
            ]
            for chunk in chunked_devices:
                fn(
                    {
                        "resource": "device",
                        "options": {"$filter": {"id": {"$in": chunk}}},
                        "body": body,
                    }
                )

    def __check_local_mode_supported(self, device: TypeDevice):
        if not is_provisioned(device):
            raise exceptions.LocalModeError(Message.DEVICE_NOT_PROVISIONED)

        if not (
            Version.parse(normalize_balena_semver(device["os_version"])) >= Version.parse(LOCAL_MODE_MIN_OS_VERSION)
        ):
            raise exceptions.LocalModeError(Message.DEVICE_OS_NOT_SUPPORT_LOCAL_MODE)

        if not (
            Version.parse(normalize_balena_semver(device["supervisor_version"]))
            >= Version.parse(LOCAL_MODE_MIN_SUPERVISOR_VERSION)
        ):
            raise exceptions.LocalModeError(Message.DEVICE_SUPERVISOR_NOT_SUPPORT_LOCAL_MODE)

        if device["os_variant"] != "dev":
            raise exceptions.LocalModeError(Message.DEVICE_OS_TYPE_NOT_SUPPORT_LOCAL_MODE)

    def __check_os_update_target(self, device_info: TypeDevice, target_os_version: str):
        if "uuid" not in device_info or not device_info["uuid"]:
            raise exceptions.OsUpdateError("The uuid of the device is not available")

        uuid = device_info["uuid"]
        if "is_online" not in device_info or not device_info["is_online"]:
            raise exceptions.OsUpdateError(f"The device is offline: {uuid}")

        if "os_version" not in device_info or not device_info["os_version"]:
            raise exceptions.OsUpdateError(f"The current os version of the device is not available: {uuid}")

        if "is_of__device_type" not in device_info or not device_info["is_of__device_type"]:
            raise exceptions.OsUpdateError(f"The device type of the device is not available: {uuid}")

        if "os_variant" not in device_info:
            raise exceptions.OsUpdateError(f"The os variant of the device is not available: {uuid}")

        current_os_version = get_device_os_semver_with_variant(device_info["os_version"], device_info["os_variant"])

        get_hup_action_type(
            device_info["is_of__device_type"][0]["slug"],
            current_os_version,
            target_os_version,
        )

    def get_dashboard_url(self, uuid: str):
        """
        Get balena Dashboard URL for a specific device.

        Args:
            uuid (str): device uuid.

        Examples:
            >>> balena.models.device.get_dashboard_url('19619a6317072b65a240b451f45f855d')
        """

        if not isinstance(uuid, str) or len(uuid) == 0:
            raise ValueError("Device UUID must be a non empty string")
        dashboard_url = cast(str, self.__settings.get("api_endpoint")).replace("api", "dashboard")
        return urljoin(dashboard_url, f"/devices/{uuid}/summary")

    def get_all(self, options: AnyObject = {}) -> List[TypeDevice]:
        """
        This method returns all devices that the current user can access.
        In order to have the following computed properties in the result
        you have to explicitly define them in a `$select` in the extra options:
         - overall_status
         - overall_progress
         - is_frozen

         Args:
            options (AnyObject): extra pine options to use

        Returns:
            List[TypeDevice]: list contains info of devices.

        Examples:
            >>> balena.models.device.get_all()
        """
        devices = self.__pine.get(
            {
                "resource": "device",
                "options": merge({"$orderby": "device_name asc"}, options),
            }
        )

        return devices

    def get_all_by_application(self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}) -> List[TypeDevice]:
        """
        Get devices by application slug, uuid or id.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[TypeDevice]: list contains info of devices.

        Examples:
            >>> balena.models.device.get_all_by_application('my_org/RPI1')
        """

        app = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})

        return self.get_all(
            merge(
                {"$filter": {"belongs_to__application": app["id"]}},
                options,
            )
        )

    def get_all_by_organization(self, handle_or_id: Union[str, int], options: AnyObject = {}) -> List[TypeDevice]:
        """
        Get devices by organization slug, uuid or id.

        Args:
            handle_or_id (Union[str, int]): organization handle (string) or id (number).
            options (AnyObject): extra pine options to use

        Returns:
            List[TypeDevice]: list contains info of devices.

        Examples:
            >>> balena.models.device.get_all_by_organization('my_org')
            >>> balena.models.device.get_all_by_organization(123)
        """
        org = self.__organization.get(handle_or_id, {"$select": "id"})
        return self.get_all(
            merge(
                {
                    "$filter": {
                        "belongs_to__application": {
                            "$any": {
                                "$alias": "bta",
                                "$expr": {"bta": {"organization": org["id"]}},
                            }
                        }
                    }
                },
                options,
            )
        )

    def get(self, uuid_or_id: Union[str, int], options: AnyObject = {}) -> TypeDevice:
        """
        This method returns a single device by id or uuid.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            options (AnyObject): extra pine options to use

        Returns:
            TypeDevice: device info.

        Examples:
            >>> balena.models.device.get('8deb12a58e3b6d3920db1c2b6303d1ff32f23d5ab99781ce1dde6876e8d143')
            >>> balena.models.device.get('8deb12')
            >>> balena.models.device.get(12345)
        """

        if uuid_or_id is None:
            raise exceptions.DeviceNotFound(uuid_or_id)

        is_potentially_full_uuid = is_full_uuid(uuid_or_id)
        if is_potentially_full_uuid or is_id(uuid_or_id):
            device = self.__pine.get(
                {
                    "resource": "device",
                    "id": {"uuid": uuid_or_id} if is_potentially_full_uuid else uuid_or_id,
                    "options": options,
                }
            )
        else:
            devices = self.__pine.get(
                {
                    "resource": "device",
                    "options": merge(
                        {"$filter": {"uuid": {"$startswith": uuid_or_id}}},
                        options,
                    ),
                }
            )

            if len(devices) > 1:
                raise exceptions.AmbiguousDevice(uuid_or_id)
            try:
                device = devices[0]
            except IndexError:
                raise exceptions.DeviceNotFound(uuid_or_id)

        if device is None:
            raise exceptions.DeviceNotFound(uuid_or_id)

        return device

    def get_with_service_details(self, uuid_or_id: Union[str, int], options: AnyObject = {}) -> TypeDeviceWithServices:
        """
        This method does not map exactly to the underlying model: it runs a
        larger prebuilt query, and reformats it into an easy to use and
        understand format. If you want more control, or to see the raw model
        directly, use `device.get(uuidOrId, options)` instead.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            options (AnyObject): extra pine options to use

        Returns:
            dict: device info with associated services details.

        Examples:
            >>> balena.models.device.get_with_service_details('0fcd753af396247e035de53b4e43eec3')
        """

        device = self.get(
            uuid_or_id,
            merge(
                {"$expand": get_current_service_details_pine_expand(True)},
                options,
            ),
        )

        return generate_current_service_details(device)

    def get_by_name(self, name: str, options: AnyObject = {}) -> List[TypeDevice]:
        """
        Get devices by device name.

        Args:
            name (str): device name.

        Returns:
            List[TypeDevice]: list contains info of devices.

        Examples:
            >>> balena.models.device.get_by_name('floral-mountain')
        """

        devices = self.get_all(merge({"$filter": {"device_name": name}}, options))

        if len(devices) == 0:
            raise exceptions.DeviceNotFound(name)

        return devices

    def get_name(self, uuid_or_id: Union[str, int]) -> str:
        """
        Get device name by device uuid.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            str: device name.

        """

        return self.get(uuid_or_id, {"$select": "device_name"})["device_name"]

    def get_application_name(self, uuid_or_id: Union[str, int]) -> str:
        """
        Get application name by device uuid.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            str: application name.
        """

        device = self.get(
            uuid_or_id,
            {
                "$select": "id",
                "$expand": {"belongs_to__application": {"$select": "app_name"}},
            },
        )
        return device["belongs_to__application"][0]["app_name"]

    def has(self, uuid_or_id: Union[str, int]) -> bool:
        """
        Check if a device exists.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            bool: True if device exists, False otherwise.
        """

        try:
            self.get(uuid_or_id, {"$select": ["id"]})
            return True
        except exceptions.DeviceNotFound:
            return False

    def is_online(self, uuid_or_id: Union[str, int]) -> bool:
        """
        Check if a device is online.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            bool: True if the device is online, False otherwise.
        """

        return self.get(uuid_or_id, {"$select": "is_online"})["is_online"]

    def get_local_ip_address(self, uuid_or_id: Union[str, int]) -> List[str]:
        """
        Get the local IP addresses of a device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            List[str]: IP addresses of a device.
        """

        device = self.get(uuid_or_id, {"$select": ["is_online", "ip_address"]})

        if not device.get("is_online"):
            raise exceptions.DeviceOffline(uuid_or_id)

        ip_address = device.get("ip_address")
        if ip_address is None:
            return []

        return ip_address.split(" ")

    def get_mac_address(self, uuid_or_id: Union[str, int]) -> List[str]:
        """
        Get the MAC addresses of a device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            List[str]: MAC addresses of a device.
        """
        device = self.get(uuid_or_id, {"$select": ["mac_address"]})

        mac = device.get("mac_address")
        if mac is None:
            return []

        return mac.split(" ")

    def get_metrics(self, uuid_or_id: Union[str, int]) -> DeviceMetricsType:
        """
        Gets the metrics related information for a device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            dict: metrics of the device.
        """

        metrics = [
            "memory_usage",
            "memory_total",
            "storage_block_device",
            "storage_usage",
            "storage_total",
            "cpu_usage",
            "cpu_temp",
            "cpu_id",
            "is_undervolted",
        ]
        device = self.get(uuid_or_id, {"$select": metrics})
        return {k: device.get(k, "") for k in metrics}

    def remove(self, uuid_or_id_or_ids: Union[str, int, List[int]]):
        """
        Remove device(s).

        Args:
            uuid_or_id_or_ids (Union[str, int, List[int]]): device uuid (str) or id (int) or ids (List[int])
        """

        self.__set(uuid_or_id_or_ids, body=None, fn=self.__pine.delete)

    def deactivate(self, uuid_or_id_or_ids: Union[str, int, List[int]]) -> None:
        """
        Deactivates a device.

        Args:
            uuid_or_id_or_ids (Union[str, int, List[int]]): device uuid (str) or id (int) or ids (List[int])

        Examples:
            >>> balena.models.device.deactivate('44cc9d1861b9f992808c506276e5d31c')
            >>> balena.models.device.deactivate([123, 234])
        """
        self.__set(uuid_or_id_or_ids, {"is_active": False})

    def rename(self, uuid_or_id: Union[str, int], new_name: str) -> None:
        """
        Renames a device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (str) or id (int)
            new_name (str): device new name.

        Examples:
            >>> balena.models.device.rename(123, 'python-sdk-test-device')
        """
        self.__set(uuid_or_id, {"device_name": new_name})

    def set_note(self, uuid_or_id_or_ids: Union[str, int, List[int]], note: str) -> None:
        """
        Note a device.

        Args:
            uuid_or_id_or_ids (Union[str, int, List[int]]): device uuid (str) or id (int) or ids (List[int])
            note (str): device note.

        Examples:
            >>> balena.models.device.note(123, 'test note')
        """
        self.__set(uuid_or_id_or_ids, {"note": note})

    def set_custom_location(
        self,
        uuid_or_id_or_ids: Union[str, int, List[int]],
        location: LocationType,
    ) -> None:
        """
        Set a custom location for a device.

        Args:
            uuid_or_id_or_ids (Union[str, int, List[int]]): device uuid (str) or id (int) or ids (List[int])
            location (dict): device custom location { 'latitude': Union[int,, str], 'longitude': Union[int, str]}.

        Examples:
            >>> balena.models.device.set_custom_location(123, {'latitude': '21.032777','longitude': '105.831586'})
        """
        self.__set(
            uuid_or_id_or_ids,
            {
                "custom_latitude": str(location["latitude"]),
                "custom_longitude": str(location["longitude"]),
            },
        )

    def unset_custom_location(self, uuid_or_id_or_ids: Union[str, int, List[int]]) -> None:
        """
        Clear the custom location of a device.

        Args:
            uuid_or_id_or_ids (Union[str, int, List[int]]): device uuid (str) or id (int) or ids (List[int])

        Examples:
            >>> balena.models.device.unset_custom_location(123)
        """

        self.set_custom_location(uuid_or_id_or_ids, {"latitude": "", "longitude": ""})

    # TODO: enable device batching
    def move(
        self,
        uuid_or_id: Union[str, int],
        app_slug_or_uuid_or_id: Union[str, int],
    ):
        """
        Move a device to another application.

        Args:
            uuid_or_id (Union[str, int]): device uuid (str) or id (int).
            app_slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number).

        Examples:
            >>> balena.models.device.move(123, 'RPI1Test')
        """
        application_options = {
            "$select": "id",
            "$expand": {
                "is_for__device_type": {
                    "$select": "is_of__cpu_architecture",
                    "$expand": {"is_of__cpu_architecture": {"$select": "slug"}},
                }
            },
        }

        app = self.__application.get(app_slug_or_uuid_or_id, application_options)
        app_cpu_arch_slug = app["is_for__device_type"][0]["is_of__cpu_architecture"][0]["slug"]

        device_options = {
            "$select": "is_of__device_type",
            "$expand": {
                "is_of__device_type": {
                    "$select": "is_of__cpu_architecture",
                    "$expand": {
                        "is_of__cpu_architecture": {
                            "$select": "slug",
                        }
                    },
                }
            },
        }

        device = self.get(uuid_or_id, device_options)
        device_cpu_arch_slug = device["is_of__device_type"][0]["is_of__cpu_architecture"][0]["slug"]

        if not self.__device_os.is_architecture_compatible_with(app_cpu_arch_slug, device_cpu_arch_slug):
            raise exceptions.IncompatibleApplication(app_slug_or_uuid_or_id)

        self.__set(uuid_or_id, {"belongs_to__application": app["id"]})

    def __supervisor_request(self, method: str, path: str, body: Optional[AnyObject] = None):
        params = {"apikey": self.__supervisor_api_key}
        req = with_supervisor_locked_error(
            lambda: requests.request(
                method=method, url=urljoin(self.__supervisor_address, path), json=body, params=params  # type: ignore
            )
        )

        if req.ok:
            try:
                return req.json()
            except JSONDecodeError:
                return req.content.decode()
        else:
            raise exceptions.RequestError(body=req.content.decode(), status_code=req.status_code)

    def __should_run_on_device(self, uuid_or_id: Optional[Union[str, int]]) -> bool:
        return self.__on_device and (uuid_or_id is None or uuid_or_id == self.__local_device_uuid)

    def ping(self, uuid_or_id: Optional[Union[str, int]] = None) -> None:
        """
        Ping a device.
        This is useful to signal that the supervisor is alive and responding.

        Args:
            uuid_or_id (Optional[Union[str, int]]): device uuid (str) or id (int) or None for SDK running on device.

        Examples:
            >>> balena.models.device.ping('8f66ec7')
            >>> balena.models.device.ping(1234)
        """

        path = "/ping"
        if self.__should_run_on_device(uuid_or_id):
            self.__supervisor_request("GET", path)
            return

        if uuid_or_id is None:
            raise exceptions.LocalSupervisorNotFound()

        device_options = {
            "$select": "id",
            "$expand": {"belongs_to__application": {"$select": "id"}},
        }

        device = self.get(uuid_or_id, device_options)
        request(
            method="POST",
            path=f"/supervisor/{path}",
            settings=self.__settings,
            body={
                "method": "GET",
                "deviceId": device["id"],
                "appId": device["belongs_to__application"][0]["id"],
            },
        )

    def identify(self, uuid_or_id: Optional[Union[str, int]] = None) -> None:
        """
        Identify device.

        Args:
            uuid_or_id (Optional[Union[str, int]]): device uuid (str) or id (int) or None for SDK running on device.

        Examples:
            >>> balena.models.device.identify('8deb12a5')
        """

        path = "/v1/blink"
        if self.__should_run_on_device(uuid_or_id):
            self.__supervisor_request("POST", path)
            return

        if uuid_or_id is None:
            raise exceptions.LocalSupervisorNotFound()

        device = self.get(uuid_or_id, {"$select": "uuid"})
        device_uuid = device["uuid"]

        request(
            method="POST",
            settings=self.__settings,
            body={"uuid": device_uuid},
            path=f"/supervisor{path}",
        )

    def restart_application(self, uuid_or_id: Optional[Union[str, int]] = None) -> None:
        """
        This function restarts the Docker container running
        the application on the device, but doesn't reboot
        the device itself.

        Args:
            uuid_or_id (Optional[Union[str, int]]): device uuid (str) or id (int) or None for SDK running on device.

        Examples:
            >>> balena.models.device.restart_application('8deb12a58')
            >>> balena.models.device.restart_application(1234)
        """

        path = "/v1/restart"
        if self.__should_run_on_device(uuid_or_id):
            self.__supervisor_request("POST", path, {"appId": self.__on_device_app_id})
            return

        if uuid_or_id is None:
            raise exceptions.LocalSupervisorNotFound()

        restart_request = {
            "$select": ["id", "supervisor_version"],
            "$expand": {"belongs_to__application": {"$select": "id"}},
        }

        def __restart_application():
            device = self.get(uuid_or_id, restart_request)
            device_id = device["id"]

            if not Version.is_valid(device["supervisor_version"]) or (
                Version.parse(device["supervisor_version"]) < Version.parse("7.0.0")
            ):
                return request(
                    method="POST",
                    path=f"device/{device_id}/restart",
                    settings=self.__settings,
                )

            app_id = device["belongs_to__application"][0]["id"]

            return request(
                method="POST",
                path=f"/supervisor{path}",
                settings=self.__settings,
                body={
                    "deviceId": device_id,
                    "appId": app_id,
                    "data": {
                        "appId": app_id,
                    },
                },
            )

        with_supervisor_locked_error(__restart_application)

    def __should_force(self, force: bool):
        if not isinstance(force, bool):
            raise ValueError(f"The `force` flag must be of type bool, got {type(force)} instead.")

        should_force = {}
        if force:
            should_force = {"force": force}

        return should_force

    def reboot(self, uuid_or_id: Optional[Union[str, int]] = None, force: bool = False) -> None:
        """
        Reboot the device.

        Args:
            uuid_or_id (Optional[Union[str, int]]): device uuid (str) or id (int) or None for SDK running on device.
            force (Optional[bool]): If force is True, the update lock will be overridden.

        Examples:
            >>> balena.models.device.reboot('8f66ec7')
        """

        path = "/v1/reboot"
        should_force = self.__should_force(force)

        if self.__should_run_on_device(uuid_or_id):
            self.__supervisor_request("POST", path, should_force)
            return

        if uuid_or_id is None:
            raise exceptions.LocalSupervisorNotFound()

        def __reboot():
            device_id = uuid_or_id if is_id(uuid_or_id) else self.get(uuid_or_id, {"$select": "id"})["id"]

            return request(
                method="POST",
                path=f"/supervisor{path}",
                settings=self.__settings,
                body={"deviceId": device_id, "data": should_force},
            )

        with_supervisor_locked_error(__reboot)

    def shutdown(self, uuid_or_id: Optional[Union[str, int]] = None, force: bool = False) -> None:
        """
        Shutdown the device.

        Args:
            uuid_or_id (Optional[Union[str, int]]): device uuid (str) or id (int) or None for SDK running on device.
            force (Optional[bool]): If force is True, the update lock will be overridden.

        Examples:
            >>> balena.models.device.shutdown('8f66ec7')
        """

        path = "/v1/shutdown"
        should_force = self.__should_force(force)

        if self.__should_run_on_device(uuid_or_id):
            self.__supervisor_request("POST", path, should_force)
            return

        if uuid_or_id is None:
            raise exceptions.LocalSupervisorNotFound()

        def __shutdown():
            device = self.get(
                uuid_or_id,
                {
                    "$select": "id",
                    "$expand": {"belongs_to__application": {"$select": "id"}},
                },
            )

            return request(
                method="POST",
                path=f"/supervisor{path}",
                settings=self.__settings,
                body={
                    "deviceId": device["id"],
                    "appId": device["belongs_to__application"][0]["id"],
                    "data": should_force,
                },
            )

        with_supervisor_locked_error(__shutdown)

    def purge(self, uuid_or_id: Optional[Union[str, int]] = None) -> None:
        """
        Purge device.
        This function clears the user application's `/data` directory.

        Args:
            uuid_or_id (Optional[Union[str, int]]): device uuid (str) or id (int) or None for SDK running on device.

        Examples:
            >>> balena.models.device.purge('8f66ec7')
        """

        path = "/v1/purge"
        if self.__should_run_on_device(uuid_or_id):
            self.__supervisor_request("POST", path, {"appId": self.__on_device_app_id})
            return

        if uuid_or_id is None:
            raise exceptions.LocalSupervisorNotFound()

        def __purge():
            device = self.get(
                uuid_or_id,
                {
                    "$select": "id",
                    "$expand": {"belongs_to__application": {"$select": "id"}},
                },
            )

            app_id = device["belongs_to__application"][0]["id"]

            return request(
                method="POST",
                path=f"/supervisor{path}",
                settings=self.__settings,
                body={
                    "deviceId": device["id"],
                    "appId": app_id,
                    "data": {"appId": app_id},
                },
            )

        with_supervisor_locked_error(__purge)

    def update(self, uuid_or_id: Optional[Union[str, int]] = None, force: bool = False) -> None:
        """
        update the device.

        Args:
            uuid_or_id (Optional[Union[str, int]]): device uuid (str) or id (int) or None for SDK running on device.
            force (Optional[bool]): If force is True, the update lock will be overridden.

        Examples:
            >>> balena.models.device.update('8f66ec7')
        """

        path = "/v1/update"
        should_force = self.__should_force(force)

        if self.__should_run_on_device(uuid_or_id):
            self.__supervisor_request("POST", path, should_force)
            return

        if uuid_or_id is None:
            raise exceptions.LocalSupervisorNotFound()

        device = self.get(
            uuid_or_id,
            {
                "$select": "id",
                "$expand": {"belongs_to__application": {"$select": "id"}},
            },
        )

        return request(
            method="POST",
            path=f"/supervisor{path}",
            settings=self.__settings,
            body={
                "deviceId": device["id"],
                "appId": device["belongs_to__application"][0]["id"],
                "data": should_force,
            },
        )

    def get_supervisor_state(self, uuid_or_id: Optional[Union[str, int]] = None) -> SupervisorStateType:
        """
        Get the supervisor state on a device

        Args:
            uuid_or_id (Optional[Union[str, int]]): device uuid (str) or id (int) or None for SDK running on device.

        Returns:
            dict: supervisor state.

        Examples:
            >>> balena.models.device.get_supervisor_state('b6070f4fea')
        """

        path = "/v1/device"
        if self.__should_run_on_device(uuid_or_id):
            return self.__supervisor_request("GET", path)  # type: ignore

        if uuid_or_id is None:
            raise exceptions.LocalSupervisorNotFound()

        uuid = self.get(uuid_or_id, {"$select": "uuid"})["uuid"]

        response = request(
            method="POST",
            path=f"/supervisor{path}",
            settings=self.__settings,
            body={"uuid": uuid, "method": "GET"},
        )

        if isinstance(response, dict):
            return response

        raise Exception(response)

    def start_service(self, uuid_or_id: Optional[Union[str, int]], image_id: int) -> None:
        """
        Start a service on device.

        Args:
            uuid_or_id (Optional[Union[str, int]]): device uuid (str) or id (int) or None for SDK running on device.
            image_id (int): id of the image to start

        Examples:
            >>> balena.models.device.start_service('f3887b1', 1234)
            >>> balena.models.device.start_service(None, 1234)  # if running on the device
        """

        if self.__should_run_on_device(uuid_or_id):
            self.__supervisor_request(
                "POST", f"/v2/applications/{self.__on_device_app_id}/start-service", {"imageId": image_id}
            )
            return

        if uuid_or_id is None:
            raise exceptions.LocalSupervisorNotFound()

        device = self.get(
            uuid_or_id,
            {
                "$select": ["id", "supervisor_version"],
                "$expand": {"belongs_to__application": {"$select": "id"}},
            },
        )
        ensure_version_compatibility(device["supervisor_version"], MIN_SUPERVISOR_MC_API, "supervisor")
        app_id = device["belongs_to__application"][0]["id"]
        request(
            method="POST",
            path=f"/supervisor/v2/applications/{app_id}/start-service",
            settings=self.__settings,
            body={
                "deviceId": device["id"],
                "appId": app_id,
                "data": {"appId": app_id, "imageId": image_id},
            },
        )

    def stop_service(self, uuid_or_id: Optional[Union[str, int]], image_id: int) -> None:
        """
        Stop a service on device.

        Args:
            uuid_or_id (Optional[Union[str, int]]): device uuid (str) or id (int) or None for SDK running on device.
            image_id (int): id of the image to stop

        Examples:
            >>> balena.models.device.stop_service('f3887b1', 392229)
            >>> balena.models.device.stop_service(None, 392229)  # if running on the device
        """

        if self.__should_run_on_device(uuid_or_id):
            self.__supervisor_request(
                "POST", f"/v2/applications/{self.__on_device_app_id}/stop-service", {"imageId": image_id}
            )
            return

        if uuid_or_id is None:
            raise exceptions.LocalSupervisorNotFound()

        def __stop_service():
            device = self.get(
                uuid_or_id,
                {
                    "$select": ["id", "supervisor_version"],
                    "$expand": {"belongs_to__application": {"$select": "id"}},
                },
            )
            ensure_version_compatibility(
                device["supervisor_version"],
                MIN_SUPERVISOR_MC_API,
                "supervisor",
            )
            app_id = device["belongs_to__application"][0]["id"]
            request(
                method="POST",
                path=f"/supervisor/v2/applications/{app_id}/stop-service",
                settings=self.__settings,
                body={
                    "deviceId": device["id"],
                    "appId": app_id,
                    "data": {"appId": app_id, "imageId": image_id},
                },
            )

        with_supervisor_locked_error(__stop_service)

    def restart_service(self, uuid_or_id: Optional[Union[str, int]], image_id: int) -> None:
        """
        Restart a service on device.

        Args:
            uuid_or_id (Optional[Union[str, int]]): device uuid (str) or id (int) or None for SDK running on device.
            image_id (int): id of the image to restart

        Examples:
            >>> balena.models.device.restart_service('f3887b', 392229)
            >>> balena.models.device.restart_service(None, 392229)  # if running on the device

        """

        if self.__should_run_on_device(uuid_or_id):
            self.__supervisor_request(
                "POST", f"/v2/applications/{self.__on_device_app_id}/restart-service", {"imageId": image_id}
            )
            return

        if uuid_or_id is None:
            raise exceptions.LocalSupervisorNotFound()

        def __restart_service():
            device = self.get(
                uuid_or_id,
                {
                    "$select": ["id", "supervisor_version"],
                    "$expand": {"belongs_to__application": {"$select": "id"}},
                },
            )
            ensure_version_compatibility(
                device["supervisor_version"],
                MIN_SUPERVISOR_MC_API,
                "supervisor",
            )
            app_id = device["belongs_to__application"][0]["id"]
            request(
                method="POST",
                path=f"/supervisor/v2/applications/{app_id}/restart_service",
                settings=self.__settings,
                body={
                    "deviceId": device["id"],
                    "appId": app_id,
                    "data": {"appId": app_id, "imageId": image_id},
                },
            )

        with_supervisor_locked_error(__restart_service)

    def get_supervisor_target_state(self, uuid_or_id: Union[str, int]) -> Any:
        """
        Get the supervisor target state on a device

        Args:
            uuid_or_id (Union[str, int]): device uuid (str) or id (int).

        Returns:
            DeviceStateType: supervisor target state.

        Examples:
            >>> balena.models.device.get_supervisor_target_state('b6070f4fea5edf808b576123157fe5ec')
        """
        device = self.get(uuid_or_id, {"$select": "uuid"})
        device_uuid = device["uuid"]

        return request(
            method="GET",
            settings=self.__settings,
            path=f"/device/v2/{device_uuid}/state",
        )

    def get_supervisor_target_state_for_app(
        self, slug_or_uuid_or_id: Union[str, int], release: Optional[Union[str, int]] = None
    ) -> Any:
        """
        Get the supervisor target state on a device

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
             release (Optional[Union[str, int]]): (optional) release uuid (default tracked)
        Returns:
            DeviceStateType: supervisor target state.

        Examples:
            >>> balena.models.device.get_supervisor_target_state_for_app('myorg/myapp')
        """
        uuid = self.__application.get(slug_or_uuid_or_id, {"$select": "uuid"})["uuid"]

        path = f"/device/v3/fleet/{uuid}/state/?releaseUuid="
        if release:
            path += str(release)

        return request(method="GET", settings=self.__settings, path=path)

    def generate_uuid(self) -> str:
        """
        Generate a random device UUID.

        Returns:
            str: a generated UUID.

        Examples:
            >>> balena.models.device.generate_uuid()
        """

        # From balena-sdk
        # I'd be nice if the UUID matched the output of a SHA-256 function,
        # but although the length limit of the CN attribute in a X.509
        # certificate is 64 chars, a 32 byte UUID (64 chars in hex) doesn't
        # pass the certificate validation in OpenVPN This either means that
        # the RFC counts a final NULL byte as part of the CN or that the
        # OpenVPN/OpenSSL implementation has a bug.
        return binascii.hexlify(os.urandom(31)).decode()

    def register(
        self,
        application_slug_or_uuid_or_id: Union[int, str],
        uuid: str,
        device_type_slug: Optional[str] = None,
    ) -> RegisterResponse:
        """
        Register a new device with a balena application.

        Args:
            application_slug_or_uuid_or_id (Union[int, str]): application slug (string), uuid (string) or id (number).
            uuid (str): device uuid.
            device_type_slug (Optional[str]): device type slug or alias.

        Returns:
            dict: dictionary contains device info.

        Examples:
            >>> device_uuid = balena.models.device.generate_uuid()
            >>> balena.models.device.register('RPI1',device_uuid)
        """

        device_type_options = {
            "$select": "slug",
            "$expand": {"is_of__cpu_architecture": {"$select": "slug"}},
        }

        application_options = {
            "$select": "id",
            "$expand": {"is_for__device_type": device_type_options},
        }

        # TODO: paralelize this 4 requests
        user_id = self.__auth.get_user_info()["id"]
        api_key = self.__application.generate_provisioning_key(application_slug_or_uuid_or_id)

        app = self.__application.get(application_slug_or_uuid_or_id, application_options)
        if isinstance(device_type_slug, str):
            device_type = self.__device_type.get(
                device_type_slug,
                {
                    "$select": "slug",
                    "$expand": {"is_of__cpu_architecture": {"$select": "slug"}},
                },
            )
        else:
            device_type = None

        if device_type is not None:
            is_compatible_parameter = self.__device_os.is_architecture_compatible_with(
                device_type["is_of__cpu_architecture"][0]["slug"],
                app["is_for__device_type"][0]["is_of__cpu_architecture"][0]["slug"],
            )

            if not is_compatible_parameter:
                app_type_slug = app["is_for__device_type"][0]["is_of__cpu_architecture"][0]["slug"]

                err_msg = f"{device_type_slug} is not compactible with application {app_type_slug} device typ"
                raise exceptions.InvalidDeviceType(err_msg)

            device_type = device_type["slug"]
        else:
            device_type = app["is_for__device_type"][0]["slug"]

        return request(
            method="POST",
            path="/device/register",
            settings=self.__settings,
            body={
                "user": user_id,
                "application": app["id"],
                "uuid": uuid,
                "device_type": device_type,
            },
            token=api_key,
        )

    # TODO: normalize response error code on no device for key response
    def generate_device_key(
        self,
        uuid_or_id: Union[str, int],
        name: Optional[str] = None,
        description: Optional[str] = None,
        expiry_date: Optional[str] = None,
    ) -> str:
        """
        Generate a device key.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            name (Optional[str]): device key name.
            description (Optional[str]): description for device key.
            expiry_date (Optional[str]): expiry date for device key, for example: `2030-01-01T00:00:00Z`.

        Examples:
            >>> balena.models.device.generate_device_key('df0926')
        """

        if is_id(uuid_or_id):
            device_id = uuid_or_id
        else:
            device_id = self.get(uuid_or_id, {"$select": "id"})["id"]

        return request(
            method="POST",
            path=f"/api-key/device/{device_id}/device-key",
            settings=self.__settings,
            body={
                "name": name,
                "description": description,
                "expiryDate": expiry_date,
            },
        )

    def has_device_url(self, uuid_or_id: Union[str, int]) -> bool:
        """
        Check if a device is web accessible with device urls

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Examples:
            >>> balena.models.device.has_device_url('8deb12a')
        """

        return self.get(uuid_or_id, {"$select": "is_web_accessible"})["is_web_accessible"]

    def get_device_url(self, uuid_or_id: Union[str, int]) -> str:
        """
        Get a device url for a device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Examples:
            >>> balena.models.device.get_device_url('8deb12a')
        """
        device = self.get(uuid_or_id, {"$select": ["uuid", "is_web_accessible"]})
        if not device["is_web_accessible"]:
            raise exceptions.DeviceNotWebAccessible(uuid_or_id)

        device_url_base = self.__config.get_all()["deviceUrlsBase"]
        uuid = device["uuid"]
        return f"https://{uuid}.{device_url_base}"

    def enable_device_url(self, uuid_or_id_or_ids: Union[str, int, List[int]]) -> None:
        """
        Enable device url for a device.

        Args:
            uuid_or_id_or_ids (Union[str, int, List[int]]): device uuid (str) or id (int) or ids (List[int]).

        Examples:
            >>> balena.models.device.enable_device_url('8deb12a58')
            >>> balena.models.device.enable_device_url([123, 345])
        """
        self.__set(uuid_or_id_or_ids, {"is_web_accessible": True})

    def disable_device_url(self, uuid_or_id_or_ids: Union[str, int, List[int]]) -> None:
        """
        Disable device url for a device.

        Args:
            uuid_or_id_or_ids (Union[str, int, List[int]]): device uuid (str) or id (int) or ids (List[int]).

        Examples:
            >>> balena.models.device.disable_device_url('8deb12a58')
            >>> balena.models.device.disable_device_url([123, 345])
        """
        self.__set(uuid_or_id_or_ids, {"is_web_accessible": False})

    def enable_local_mode(self, uuid_or_id: Union[str, int]) -> None:
        """
        Enable local mode.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Examples:
            >>> balena.models.device.enable_local_mode('b6070f4f')
        """

        device = self.get(uuid_or_id, {"$select": self.__LOCAL_MODE_SELECT})
        self.__check_local_mode_supported(device)
        self.config_var.set(uuid_or_id, LOCAL_MODE_ENV_VAR, "1")

    def disable_local_mode(self, uuid_or_id: Union[str, int]) -> None:
        """
        Disable local mode.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            None.

        Examples:
            >>> balena.models.device.disable_local_mode('b6070f4f')
        """
        self.config_var.set(uuid_or_id, LOCAL_MODE_ENV_VAR, "0")

    def is_in_local_mode(self, uuid_or_id: Union[str, int]) -> bool:
        """
        Check if local mode is enabled on the device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            bool: True if local mode enabled, otherwise False.

        Examples:
            >>> balena.models.device.is_in_local_mode('b6070f4f')
        """

        device = self.get(uuid_or_id, {"$select": "id"})
        result = self.__pine.get(
            {
                "resource": "device_config_variable",
                "id": {"device": device["id"], "name": LOCAL_MODE_ENV_VAR},
            }
        )

        return result is not None and result.get("value") == "1"

    def get_local_mode_support(self, uuid_or_id: Union[str, int]) -> LocalModeResponse:
        """
        Returns whether local mode is supported and a message describing the reason why local mode is not supported.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            dict: local mode support information ({'supported': True/False, 'message': '...'}).

        Examples:
            >>> balena.models.device.get_local_mode_support('b6070f4')
        """

        device = self.get(uuid_or_id, {"$select": self.__LOCAL_MODE_SELECT})
        try:
            self.__check_local_mode_supported(device)
            return {"supported": True, "message": "Supported"}
        except exceptions.LocalModeError as e:
            return {"supported": False, "message": e.message}

    def enable_lock_override(self, uuid_or_id: Union[str, int]) -> None:
        """
        Enable lock override.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
        """
        self.config_var.set(uuid_or_id, OVERRIDE_LOCK_ENV_VAR, "1")

    def disable_lock_override(self, uuid_or_id: Union[str, int]) -> None:
        """
        Disable lock override.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
        """
        self.config_var.set(uuid_or_id, OVERRIDE_LOCK_ENV_VAR, "0")

    def has_lock_override(self, uuid_or_id: Union[str, int]) -> bool:
        """
        Check if a device has the lock override enabled.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            bool: lock override status.
        """

        return self.__get_applied_device_config_variable_value(uuid_or_id, OVERRIDE_LOCK_ENV_VAR) == "1"

    def get_status(self, uuid_or_id: Union[str, int]) -> str:
        """
        Get the status of a device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            str: status of a device.

        Examples:
            >>> balena.models.device.get_status('8deb12')
        """

        return self.get(uuid_or_id, {"$select": "overall_status"})["overall_status"]

    def grant_support_access(
        self,
        uuid_or_id_or_ids: Union[str, int, List[int]],
        expiry_timestamp: int,
    ) -> None:
        """
        Grant support access to a device until a specified time.

        Args:
            uuid_or_id_or_ids (Union[str, int, List[int]]): device uuid (str) or id (int) or ids (List[int])
            expiry_timestamp (int): a timestamp in ms for when the support access will expire.

        Examples:
            >>> balena.models.device.grant_support_access('49b2a7', 1511974999000)
        """

        if expiry_timestamp is None or expiry_timestamp <= int(
            (datetime.datetime.utcnow() - datetime.datetime.utcfromtimestamp(0)).total_seconds() * 1000
        ):
            raise exceptions.InvalidParameter("expiry_timestamp", expiry_timestamp)

        self.__set(
            uuid_or_id_or_ids,
            {"is_accessible_by_support_until__date": expiry_timestamp},
        )

    def revoke_support_access(self, uuid_or_id_or_ids: Union[str, int, List[int]]) -> None:
        """
        Revoke support access to a device.

        Args:
            uuid_or_id_or_ids (Union[str, int, List[int]]): device uuid (str) or id (int) or ids (List[int])

        Examples:
            >>> balena.models.device.revoke_support_access('49b2a7')
        """
        self.__set(
            uuid_or_id_or_ids,
            {"is_accessible_by_support_until__date": None},
        )

    def is_tracking_application_release(self, uuid_or_id: Union[str, int]) -> bool:
        """
        Get whether the device is configured to track the current application release.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Returns:
            bool: is tracking the current application release.
        """

        return not bool(self.get(uuid_or_id, {"$select": "should_be_running__release"})["should_be_running__release"])

    # TODO: enable device batching
    def pin_to_release(
        self,
        uuid_or_id: Union[str, int],
        full_release_hash_or_id: Union[str, int],
    ) -> None:
        """
        Configures the device to run a particular release
        and not get updated when the current application release changes.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            full_release_hash_or_id (Union[str, int]) : the hash of a successful release (string) or id (number)

        Examples:
            >>> balena.models.device.set_to_release('49b2a', '45c90004de73557ded7274d4896a6db90ea61e36')
        """

        device = self.get(
            uuid_or_id,
            {
                "$select": "id",
                "$expand": {"belongs_to__application": {"$select": "id"}},
            },
        )
        app_id = device["belongs_to__application"][0]["id"]
        release_options = {
            "$top": 1,
            "$select": "id",
            "$filter": {
                "status": "success",
                "belongs_to__application": app_id,
            },
            "$orderby": "created_at desc",
        }
        if is_id(full_release_hash_or_id):
            release_options["$filter"]["id"] = full_release_hash_or_id
        else:
            release_options["$filter"]["commit"] = full_release_hash_or_id

        release = self.__release.get(full_release_hash_or_id, release_options)
        self.__pine.patch(
            {
                "resource": "device",
                "id": device["id"],
                "body": {"should_be_running__release": release["id"]},
            }
        )

    def track_application_release(self, uuid_or_id_or_ids: Union[str, int, List[int]]) -> None:
        """
        Configure a specific device to track the current application release.

        Args:
            uuid_or_id_or_ids (Union[str, int, List[int]]): device uuid (str) or id (int) or ids (List[int])
        """

        self.__set(uuid_or_id_or_ids, {"should_be_running__release": None})

    # TODO: enable device batching
    def set_supervisor_release(
        self,
        uuid_or_id: Union[str, int],
        supervisor_version_or_id: Union[str, int],
    ) -> None:
        """
        Set a specific device to run a particular supervisor release.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            supervisor_version_or_id (Union[str, int]): the version of a released supervisor (string) or id (number)

        Examples:
            >>> balena.models.device.set_supervisor_release('f55dcdd9ad', 'v13.0.0')
        """
        device = self.get(
            uuid_or_id,
            {
                "$select": ["id", "supervisor_version", "os_version"],
                "$expand": {"is_of__device_type": {"$select": "slug"}},
            },
        )
        device_type_slug = device["is_of__device_type"][0]["slug"]

        release_options = {
            "$top": 1,
            "$select": "id",
            "$filter": {
                "is_for__device_type": {
                    "$any": {
                        "$alias": "dt",
                        "$expr": {
                            "dt": {
                                "slug": device_type_slug,
                            },
                        },
                    },
                },
            },
        }

        if is_id(supervisor_version_or_id):
            release_options["$filter"]["id"] = supervisor_version_or_id
        else:
            release_options["$filter"]["supervisor_version"] = supervisor_version_or_id

        try:
            release = self.__pine.get({"resource": "supervisor_release", "options": release_options})[0]
        except IndexError:
            raise Exception(f"Supervisor release not found {supervisor_version_or_id}")

        ensure_version_compatibility(device["supervisor_version"], MIN_SUPERVISOR_MC_API, "supervisor")
        ensure_version_compatibility(device["os_version"], MIN_OS_MC, "host OS")

        self.__pine.patch(
            {
                "resource": "device",
                "id": device["id"],
                "body": {"should_be_managed_by__supervisor_release": release["id"]},
            }
        )

    def start_os_update(self, uuid_or_id: Union[str, int], target_os_version: str) -> HUPStatusResponse:
        """
        Start an OS update on a device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int).
            target_os_version (str): semver-compatible version for the target device.
                Unsupported (unpublished) version will result in rejection.
                The version **must** be the exact version number, a "prod" variant
                and greater than the one running on the device.

        Returns:
            HUPStatusResponse: action response.

        Examples:
            >>> balena.models.device.start_os_update('b6070f4', '2.29.2+rev1.prod')
            >>> balena.models.device.start_os_update('b6070f4', '2.89.0+rev1')
        """

        if target_os_version is None or uuid_or_id is None:
            raise exceptions.InvalidParameter("target_os_version or UUID", None)

        device = self.get(
            uuid_or_id,
            {
                "$select": ["uuid", "is_online", "os_version", "os_variant"],
                "$expand": {"is_of__device_type": {"$select": "slug"}},
            },
        )

        self.__check_os_update_target(device, target_os_version)

        all_versions = self.__device_os.get_available_os_versions(device["is_of__device_type"][0]["slug"])
        if not [v for v in all_versions if target_os_version == v["raw_version"]]:
            raise exceptions.InvalidParameter("target_os_version", target_os_version)

        data = {"parameters": {"target_version": target_os_version}}

        url_base = self.__config.get_all()["deviceUrlsBase"]
        action_api_version = self.__settings.get("device_actions_endpoint_version")

        return request(
            method="POST",
            settings=self.__settings,
            path=f"{device['uuid']}/{self.__device_os.OS_UPDATE_ACTION_NAME}",
            body=data,
            endpoint=f"https://actions.{url_base}/{action_api_version}/",
        )

    def get_os_update_status(self, uuid_or_id: Union[str, int]) -> HUPStatusResponse:
        """
        Get the OS update status of a device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int).

        Returns:
            HUPStatusResponse: action response.

        Examples:
            >>> balena.models.device.get_os_update_status('b6070f4f')
        """

        device = self.get(uuid_or_id, {"$select": "uuid"})
        url_base = self.__config.get_all()["deviceUrlsBase"]
        action_api_version = self.__settings.get("device_actions_endpoint_version")

        return request(
            method="GET",
            settings=self.__settings,
            path=f"{device['uuid']}/{self.__device_os.OS_UPDATE_ACTION_NAME}",
            endpoint=f"https://actions.{url_base}/{action_api_version}/",
        )

    @deprecated("This is not supported on multicontainer devices, and will be removed in a future major release")
    def get_application_info(self, uuid_or_id: Union[str, int]) -> Any:
        """
        ***Deprecated***
        Return information about the application running on the device.
        This function requires supervisor v1.8 or higher.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int).

        Returns:
            dict: dictionary contains application information.

        Examples:
            >>> balena.models.device.get_application_info('7f66ec')
        """

        device = self.get(
            uuid_or_id,
            {
                "$select": ["id", "supervisor_version"],
                "$expand": {"belongs_to__application": {"$select": "id"}},
            },
        )

        ensure_version_compatibility(device["supervisor_version"], MIN_SUPERVISOR_APPS_API, "supervisor")
        app_id = device["belongs_to__application"][0]["id"]

        return request(
            method="POST",
            path=f"/supervisor/v1/apps/{app_id}",
            settings=self.__settings,
            body={"deviceId": device["id"], "appId": app_id, "method": "GET"},
        )

    @deprecated("This is not supported on multicontainer devices, and will be removed in a future major release")
    def start_application(self, uuid_or_id: Union[str, int]) -> None:
        """
        ***Deprecated***
        Starts a user application container, usually after it has been stopped with `stop_application()`.
        This function requires supervisor v1.8 or higher.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int).

        Returns:
            dict: dictionary contains started application container id.

        Examples:
            >>> balena.models.device.start_application('8f66ec7')
        """

        device = self.get(
            uuid_or_id,
            {
                "$select": ["id", "supervisor_version"],
                "$expand": {"belongs_to__application": {"$select": "id"}},
            },
        )

        ensure_version_compatibility(device["supervisor_version"], MIN_SUPERVISOR_APPS_API, "supervisor")
        app_id = device["belongs_to__application"][0]["id"]
        request(
            method="POST",
            path=f"/supervisor/v1/apps/{app_id}/start",
            settings=self.__settings,
            body={
                "deviceId": device["id"],
                "appId": app_id,
            },
        )

    @deprecated("This is not supported on multicontainer devices, and will be removed in a future major release")
    def stop_application(self, uuid_or_id: Union[str, int]):
        """
        ***Deprecated***
        Temporarily stops a user application container.
        Application container will not be removed after invoking this function and a
        reboot or supervisor restart will cause the container to start again.
        This function requires supervisor v1.8 or higher.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int).

        Returns:
            dict: dictionary contains stopped application container id.

        Examples:
            >>> balena.models.device.stop_application('8f66ec')
        """

        def __stop_aplication():
            device = self.get(
                uuid_or_id,
                {
                    "$select": ["id", "supervisor_version"],
                    "$expand": {"belongs_to__application": {"$select": "id"}},
                },
            )

            ensure_version_compatibility(
                device["supervisor_version"],
                MIN_SUPERVISOR_APPS_API,
                "supervisor",
            )
            app_id = device["belongs_to__application"][0]["id"]
            request(
                method="POST",
                path=f"/supervisor/v1/apps/{app_id}/stop",
                settings=self.__settings,
                body={
                    "deviceId": device["id"],
                    "appId": app_id,
                },
            )

        with_supervisor_locked_error(__stop_aplication)


class DeviceTag(DependentResource[BaseTagType]):
    """
    This class implements device tag model for balena python SDK.

    """

    def __init__(self, pine: PineClient, device: Device, application: Application):
        self.__device = device
        self.__application = application
        super(DeviceTag, self).__init__(
            "device_tag", "tag_key", "device", lambda id: self.__device.get(id, {"$select": "id"})["id"], pine
        )

    def get_all_by_application(self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}) -> List[BaseTagType]:
        """
        Get all device tags for an application.

        Args:
            slug_or_uuid_or_id (int): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[BaseTagType]: tags list.

        Examples:
            >>> balena.models.device.tag.get_all_by_application(1005160)
        """

        app_id = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})["id"]
        return super(DeviceTag, self)._get_all(
            merge(
                {
                    "$filter": {
                        "device": {
                            "$any": {
                                "$alias": "d",
                                "$expr": {"d": {"belongs_to__application": app_id}},
                            }
                        }
                    }
                },
                options,
            )
        )

    def get_all_by_device(self, uuid_or_id: Union[str, int], options: AnyObject = {}) -> List[BaseTagType]:
        """
        Get all device tags for a device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[BaseTagType]: tags list.

        Examples:
            >>> balena.models.device.tag.get_all_by_device('a03ab646c')
        """

        id = self.__device.get(uuid_or_id, {"$select": "id"})["id"]
        return super(DeviceTag, self)._get_all_by_parent(id, options)

    def get_all(self, options: AnyObject = {}) -> List[BaseTagType]:
        """
        Get all device tags.

        Args:
            options (AnyObject): extra pine options to use

        Returns:
            List[BaseTagType]: tags list.

        Examples:
            >>> balena.models.device.tag.get_all()
        """

        return super(DeviceTag, self)._get_all(options)

    def get(self, uuid_or_id: Union[str, int], tag_key: str) -> Optional[str]:
        """
        Set a device tag (update tag value if it exists).
        ___Note___: Notice that when using the device ID rather than UUID,
        it will avoid one extra API roundtrip.

        Args:
            uuid_or_id (Union[str, int]): device uuid or device id.
            tag_key (str): tag key.
            value (str): tag value.

        Returns:
            Optional[str]: tag value

        Examples:
            >>> balena.models.device.tag.set('f5213eac0d63ac4', 'testtag','test1')
            >>> balena.models.device.tag.set('f5213eac0d63ac4', 'testtag','test2')
        """

        # Trying to avoid an extra HTTP request
        # when the provided parameter looks like an id.
        # Note that this throws an exception for missing names/uuids,
        # but not for missing ids
        device_id = uuid_or_id if is_id(uuid_or_id) else self.__device.get(uuid_or_id, {"$select": "id"})["id"]
        return super(DeviceTag, self)._get(device_id, tag_key)

    def set(self, uuid_or_id: Union[str, int], tag_key: str, value: str) -> None:
        """
        Set a device tag (update tag value if it exists).
        ___Note___: Notice that when using the device ID rather than UUID,
        it will avoid one extra API roundtrip.

        Args:
            uuid_or_id (Union[str, int]): device uuid or device id.
            tag_key (str): tag key.
            value (str): tag value.

        Examples:
            >>> balena.models.device.tag.set('f5213eac0d63ac4', 'testtag','test1')
            >>> balena.models.device.tag.set('f5213eac0d63ac4', 'testtag','test2')
        """

        # Trying to avoid an extra HTTP request
        # when the provided parameter looks like an id.
        # Note that this throws an exception for missing names/uuids,
        # but not for missing ids
        device_id = uuid_or_id if is_id(uuid_or_id) else self.__device.get(uuid_or_id, {"$select": "id"})["id"]
        super(DeviceTag, self)._set(device_id, tag_key, value)

    def remove(self, uuid_or_id: Union[str, int], tag_key: str) -> None:
        """
        Remove a device tag.

        Args:
            uuid_or_id (Union[str, int]): device uuid or device id.
            tag_key (str): tag key.

        Examples:
            >>> balena.models.device.tag.remove('f5213eac0d63ac477', 'testtag')
        """

        device_id = uuid_or_id if is_id(uuid_or_id) else self.__device.get(uuid_or_id, {"$select": "id"})["id"]
        super(DeviceTag, self)._remove(device_id, tag_key)


class DeviceConfigVariable(DependentResource[EnvironmentVariableBase]):
    """
    This class implements device config variable model for balena python SDK.

    """

    def __init__(self, pine: PineClient, device: Device, application: Application):
        self.__device = device
        self.__application = application
        super(DeviceConfigVariable, self).__init__(
            "device_config_variable", "name", "device", lambda id: self.__device.get(id, {"$select": "id"})["id"], pine
        )

    def get_all_by_device(self, uuid_or_id: Union[str, int], options: AnyObject = {}) -> List[EnvironmentVariableBase]:
        """
        Get all device config variables belong to a device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            options (AnyObject): extra pine options to use

        Returns:
            List[EnvironmentVariableBase]: device config variables.

        Examples:
            >>> balena.models.device.config_var.get_all_by_device('f5213ea')
        """
        return super(DeviceConfigVariable, self)._get_all_by_parent(uuid_or_id, options)

    def get_all_by_application(
        self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[EnvironmentVariableBase]:
        """
        Get all device config variables for an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[EnvironmentVariableBase]: list of device environment variables.

        Examples:
            >>> balena.models.device.config_var.device.get_all_by_application(5780)
        """
        app_id = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})["id"]
        return super(DeviceConfigVariable, self)._get_all(
            merge(
                {
                    "$filter": {
                        "device": {
                            "$any": {
                                "$alias": "d",
                                "$expr": {
                                    "d": {
                                        "belongs_to__application": app_id,
                                    },
                                },
                            },
                        },
                    },
                    "$orderby": "name asc",
                },
                options,
            )
        )

    def get(self, uuid_or_id: Union[str, int], env_var_name: str) -> Optional[str]:
        """
        Get a device config variable.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            env_var_name (str): environment variable name.

        Examples:
            >>> balena.models.device.config_var.device.get('8deb12','test_env4')
        """
        return super(DeviceConfigVariable, self)._get(uuid_or_id, env_var_name)

    def set(self, uuid_or_id: Union[str, int], env_var_name: str, value: str) -> None:
        """
        Set the value of a device config variable.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            env_var_name (str): environment variable name.
            value (str): environment variable value.

        Examples:
            >>> balena.models.device.config_var.device.set('8deb12','test_env4', 'testing1')
        """
        super(DeviceConfigVariable, self)._set(uuid_or_id, env_var_name, value)

    def remove(self, uuid_or_id: Union[str, int], key: str) -> None:
        """
        Remove a device environment variable.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Examples:
            >>> balena.models.device.config_var.device.remove(2184)
        """
        super(DeviceConfigVariable, self)._remove(uuid_or_id, key)


class DeviceEnvVariable(DependentResource[EnvironmentVariableBase]):
    """
    This class implements device environment variable model for balena python SDK.

    """

    def __init__(self, pine: PineClient, device: Device, application: Application):
        self.__device = device
        self.__application = application
        super(DeviceEnvVariable, self).__init__(
            "device_environment_variable",
            "name",
            "device",
            lambda id: self.__device.get(id, {"$select": "id"})["id"],
            pine,
        )

    def get_all_by_device(self, uuid_or_id: Union[str, int], options: AnyObject = {}) -> List[EnvironmentVariableBase]:
        """
        Get all device environment variables.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            options (AnyObject): extra pine options to use

        Returns:
            List[EnvironmentVariableBase]: device environment variables.

        Examples:
            >>> balena.models.device.env_var.get_all_by_device('8deb12a')
        """
        return super(DeviceEnvVariable, self)._get_all_by_parent(uuid_or_id, options)

    def get_all_by_application(
        self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[EnvironmentVariableBase]:
        """
        Get all device environment variables for an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[EnvironmentVariableBase]: list of device environment variables.

        Examples:
            >>> balena.models.device.env_var.get_all_by_application(5780)
        """
        app_id = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})["id"]
        return super(DeviceEnvVariable, self)._get_all(
            merge(
                {
                    "$filter": {
                        "device": {
                            "$any": {
                                "$alias": "d",
                                "$expr": {
                                    "d": {
                                        "belongs_to__application": app_id,
                                    },
                                },
                            },
                        },
                    },
                    "$orderby": "name asc",
                },
                options,
            )
        )

    def get(self, uuid_or_id: Union[str, int], env_var_name: str) -> Optional[str]:
        """
        Get device environment variable.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            env_var_name (str): environment variable name.

        Examples:
            >>> balena.models.device.env_var.get('8deb12','test_env4')
        """
        return super(DeviceEnvVariable, self)._get(uuid_or_id, env_var_name)

    def set(self, uuid_or_id: Union[str, int], env_var_name: str, value: str) -> None:
        """
        Set the value of a specific environment variable.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            env_var_name (str): environment variable name.
            value (str): environment variable value.

        Examples:
            >>> balena.models.device.env_var.set('8deb12','test_env4', 'testing1')
        """
        super(DeviceEnvVariable, self)._set(uuid_or_id, env_var_name, value)

    def remove(self, uuid_or_id: Union[str, int], key: str) -> None:
        """
        Remove a device environment variable.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)

        Examples:
            >>> balena.models.device.env_var.remove(2184)
        """
        super(DeviceEnvVariable, self)._remove(uuid_or_id, key)


class DeviceServiceEnvVariable:
    """
    This class implements device service variable model for balena python SDK.
    """

    def __init__(self, pine: PineClient, device: Device, application: Application):
        self.__pine = pine
        self.__device = device
        self.__application = application

    def get_all_by_device(self, uuid_or_id: Union[str, int], options: AnyObject = {}) -> List[EnvironmentVariableBase]:
        """
        Get all device environment variables.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            options (AnyObject): extra pine options to use

        Returns:
            List[EnvironmentVariableBase]: device service environment variables.

        Examples:
            >>> balena.models.device.service_var.get_all_by_device(8deb12a)
        """
        device_id = self.__device.get(uuid_or_id, {"$select": "id"})["id"]
        return self.__pine.get(
            {
                "resource": "device_service_environment_variable",
                "options": merge(
                    {
                        "$filter": {
                            "service_install": {
                                "$any": {
                                    "$alias": "si",
                                    "$expr": {"si": {"device": device_id}},
                                }
                            }
                        }
                    },
                    options,
                ),
            }
        )

    def get_all_by_application(
        self, slug_or_uuid_or_id: Union[str, int], options: AnyObject = {}
    ) -> List[EnvironmentVariableBase]:
        """
        Get all device service environment variables belong to an application.

        Args:
            slug_or_uuid_or_id (Union[str, int]): application slug (string), uuid (string) or id (number)
            options (AnyObject): extra pine options to use

        Returns:
            List[EnvironmentVariableBase]: device service environment variables.

        Examples:
            >>> balena.models.device.service_var.get_all_by_application(1043050)
        """
        app_id = self.__application.get(slug_or_uuid_or_id, {"$select": "id"})["id"]

        return self.__pine.get(
            {
                "resource": "device_service_environment_variable",
                "options": merge(
                    {
                        "$filter": {
                            "service_install": {
                                "$any": {
                                    "$alias": "si",
                                    "$expr": {
                                        "si": {
                                            "device": {
                                                "$any": {
                                                    "$alias": "d",
                                                    "$expr": {"d": {"belongs_to__application": app_id}},
                                                }
                                            }
                                        }
                                    },
                                }
                            }
                        },
                        "$orderby": "name asc",
                    },
                    options,
                ),
            }
        )

    def get(self, uuid_or_id: Union[str, int], service_name_or_id: Union[str, int], key: str) -> Optional[str]:
        """
        Get the overriden value of a service variable on a device

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            service_name_or_id (Union[str, int]): service name (string) or service id (number)
            key (str): variable name

        Returns:
           Optional[str]: device service environment variables.

        Examples:
            >>> balena.models.device.service_var.get('8deb12a', 'myservice', 'VAR')
            >>> balena.models.device.service_var.get('8deb12a', 1234', 'VAR')
        """
        device_id = self.__device.get(uuid_or_id, {"$select": "id"})["id"]

        installs_service = (
            service_name_or_id
            if isinstance(service_name_or_id, int)
            else {"$any": {"$alias": "is", "$expr": {"is": {"service_name": service_name_or_id}}}}
        )

        variables = self.__pine.get(
            {
                "resource": "device_service_environment_variable",
                "options": {
                    "$select": "value",
                    "$filter": {
                        "service_install": {
                            "$any": {
                                "$alias": "si",
                                "$expr": {"si": {"device": device_id, "installs__service": installs_service}},
                            }
                        },
                        "name": key,
                    },
                },
            }
        )

        if isinstance(variables, list) and len(variables) == 1:
            return variables[0].get("value")

    def set(self, uuid_or_id: Union[str, int], service_name_or_id: Union[str, int], key: str, value: str) -> None:
        """
        Set the overriden value of a service variable on a device.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            service_name_or_id (Union[str, int]): service name (string) or service id (number)
            key (str): variable name
            value (str): variable value

        Examples:
            >>> balena.models.device.service_var.set('7cf02a6', 'myservice', 'VAR', 'override')
            >>> balena.models.device.service_var.set('7cf02a6', 123, 'VAR', 'override')
        """

        if is_id(uuid_or_id):
            device_filter = uuid_or_id
        elif is_full_uuid(uuid_or_id):
            device_filter = {"$any": {"$alias": "d", "$expr": {"d": {"uuid": uuid_or_id}}}}
        else:
            device_filter = self.__device.get(uuid_or_id, {"$select": "id"})["id"]

        installs_service = (
            service_name_or_id
            if isinstance(service_name_or_id, int)
            else {"$any": {"$alias": "s", "$expr": {"s": {"service_name": service_name_or_id}}}}
        )

        service_installs = self.__pine.get(
            {
                "resource": "service_install",
                "options": {
                    "$select": "id",
                    "$filter": {
                        "device": device_filter,
                        "installs__service": installs_service,
                    },
                },
            }
        )

        if (
            service_installs is None
            or (isinstance(service_installs, list) and len(service_installs) == 0)
            or service_installs[0] is None
        ):
            raise exceptions.ServiceNotFound(service_name_or_id)

        if len(service_installs) > 1:
            raise exceptions.AmbiguousDevice(uuid_or_id)

        self.__pine.upsert(
            {
                "resource": "device_service_environment_variable",
                "id": {
                    "service_install": service_installs[0]["id"],
                    "name": key,
                },
                "body": {"value": value},
            }
        )

    def remove(self, uuid_or_id: Union[str, int], service_name_or_id: Union[str, int], key: str) -> None:
        """
        Remove a device service environment variable.

        Args:
            uuid_or_id (Union[str, int]): device uuid (string) or id (int)
            service_name_or_id (Union[str, int]): service name (string) or service id (number)
            key (str): variable name

        Examples:
            >>> balena.models.device.service_var.set('7cf02a6', 'myservice', 'VAR')
            >>> balena.models.device.service_var.remove('7cf02a6', 28970, 'VAR')
        """

        device_id = self.__device.get(uuid_or_id, {"$select": "id"})["id"]
        installs_service = (
            service_name_or_id
            if isinstance(service_name_or_id, int)
            else {"$any": {"$alias": "is", "$expr": {"is": {"service_name": service_name_or_id}}}}
        )

        self.__pine.delete(
            {
                "resource": "device_service_environment_variable",
                "options": {
                    "$filter": {
                        "service_install": {
                            "$any": {
                                "$alias": "si",
                                "$expr": {
                                    "si": {
                                        "device": device_id,
                                        "installs__service": installs_service,
                                    }
                                },
                            }
                        },
                        "name": key,
                    }
                },
            }
        )



"""
This module implements all models for balena python SDK.

"""

from ..pine import PineClient
from .api_key import ApiKey
from .application import Application
from .billing import Billing
from .credit_bundle import CreditBundle
from .config import Config
from .device import Device
from .device_type import DeviceType
from .image import Image
from .key import Key
from .organization import Organization
from .os import DeviceOs
from .release import Release
from .service import Service
from ..settings import Settings


class Models:
    def __init__(self, pine: PineClient, settings: Settings):
        self.application = Application(pine, settings)
        self.billing = Billing(pine, settings)
        self.credit_bundle = CreditBundle(pine, settings)
        self.device = Device(pine, settings)
        self.device_type = DeviceType(pine, settings)
        self.api_key = ApiKey(pine, settings)
        self.key = Key(pine, settings)
        self.organization = Organization(pine, settings)
        self.os = DeviceOs(pine, settings)
        self.config = Config(settings)
        self.release = Release(pine, settings)
        self.service = Service(pine, settings)
        self.image = Image(pine, settings)



from . import exceptions
from .balena_auth import request
from .settings import Settings
from typing import TypedDict, Optional, Literal, Union, cast
from typing_extensions import Unpack
from .pine import PineClient
from .twofactor_auth import TwoFactorAuth


TOKEN_KEY = "token"


class CredentialsType(TypedDict):
    username: str
    password: str


class UserKeyWhoAmIResponse(TypedDict):
    id: int
    actorType: Literal["user"]
    actorTypeId: int
    username: str
    email: Optional[str]


class ApplicationKeyWhoAmIResponse(TypedDict):
    id: int
    actorType: Literal["application"]
    actorTypeId: int
    slug: str


class DeviceKeyWhoAmIResponse(TypedDict):
    id: int
    actorType: Literal["device"]
    actorTypeId: int
    uuid: str


WhoamiResult = Union[UserKeyWhoAmIResponse, ApplicationKeyWhoAmIResponse, DeviceKeyWhoAmIResponse]


class UserInfo(TypedDict):
    id: int
    actor: int
    username: str
    email: Optional[str]


class Auth:
    """
    This class implements all authentication functions for balena python SDK.

    """

    _actor_details_cache: Optional[WhoamiResult] = None
    _user_actor_id_cache: Optional[int] = None

    def __init__(self, pine: PineClient, settings: Settings):
        self.two_factor = TwoFactorAuth(settings)
        self.__pine = pine
        self.__settings = settings

    def __get_actor_details(self, no_cache: bool = False) -> WhoamiResult:
        """
        Get user details from token.

        Returns:
            Optional[WhoamiResult]: user details.
        """
        if not self._actor_details_cache or no_cache:
            whoami = request(method="GET", settings=self.__settings, path="/actor/v1/whoami")
            if isinstance(whoami, dict) and set(["id", "actorType"]).issubset(set(whoami.keys())):
                self._actor_details_cache = cast(WhoamiResult, whoami)
            else:
                raise exceptions.NotLoggedIn()

        return self._actor_details_cache

    def whoami(self) -> Optional[WhoamiResult]:
        """
        Return current logged in username.

        Returns:
            Optional[WhoamiResult]: current logged in information

        Examples:
            >>> balena.auth.whoami()
        """
        return self.__get_actor_details()

    def authenticate(self, **credentials: Unpack[CredentialsType]) -> str:
        """
        This function authenticates provided credentials information.
        You should use Auth.login when possible, as it takes care of saving the Auth Token and username as well.

        Args:
            **credentials: credentials keyword arguments.
                username (str): Balena username.
                password (str): Password.

        Returns:
            str: Auth Token,

        Examples:
            >>> balena.auth.authenticate(username='<your email>', password='<your password>')
        """
        req = request(
            method="POST", settings=self.__settings, path="login_", body=credentials, send_token=False, return_raw=True
        )

        if not req.ok:
            if req.status_code == 401:
                raise exceptions.LoginFailed()
            elif req.status_code == 429:
                raise exceptions.TooManyRequests()

        return req.content.decode()

    def login(self, **credentials: Unpack[CredentialsType]) -> None:
        """
        This function is used for logging into balena using email and password.

        Args:
            **credentials: credentials keyword arguments.
                username (str): Balena email.
                password (str): Password.

        Examples:
            >>> from balena import Balena
            ... balena = Balena()
            ... credentials = {'username': '<your email>', 'password': '<your password>'}
            ... balena.auth.login(**credentials)
            ... # or
            ... balena.auth.login(username='<your email>', password='<your password>')
        """
        token = self.authenticate(**credentials)
        self._actor_details_cache = None
        self._user_actor_id_cache = None
        self.__settings.set(TOKEN_KEY, token)

    def login_with_token(self, token: str) -> None:
        """
        This function is used for logging into balena using Auth Token.
        Auth Token can be found in Preferences section on balena Dashboard.

        Args:
            token (str): Auth Token.

        Returns:
            This functions saves Auth Token to Settings and returns nothing.

        Examples:
            >>> from balena import Balena
            >>> balena = Balena()
            >>> auth_token = <your token>
            >>> balena.auth.login_with_token(auth_token)

        """
        self._actor_details_cache = None
        self._user_actor_id_cache = None
        self.__settings.set(TOKEN_KEY, token)

    def is_logged_in(self) -> bool:
        """
        This function checks if you're logged in

        Returns:
            bool: True if logged in, False otherwise.

        Examples:
            # Check if user logged in.
            >>> if balena.auth.is_logged_in():
            ...     print('You are logged in!')
            ... else:
            ...     print('You are not logged in!')
        """
        try:
            self.__get_actor_details(True)
            return True
        except (
            exceptions.RequestError,
            exceptions.Unauthorized,
            exceptions.NotLoggedIn,
        ):
            return False

    def get_token(self) -> Optional[str]:
        """
        This function retrieves Auth Token.

        Returns:
            str: Auth Token.

        Examples:
            >>> balena.auth.get_token()
        """
        try:
            return cast(str, self.__settings.get(TOKEN_KEY))
        except exceptions.InvalidOption:
            return None

    def get_user_info(self) -> UserInfo:
        """
        Get current logged in user's info

        Returns:
            UserInfo: user info.

        Examples:
            # If you are logged in as a user.
            >>> balena.auth.get_user_info()
        """
        actor = self.__get_actor_details()
        if actor and actor["actorType"] != "user":
            raise Exception("The authentication credentials in use are not of a user")

        actor = cast(UserKeyWhoAmIResponse, actor)
        return {
            "id": actor["actorTypeId"],
            "actor": actor["id"],
            "email": actor["email"],
            "username": actor["username"],
        }

    def get_actor_id(self) -> int:
        """
        Get current logged in actor id.

        Returns:
            int: actor id

        Examples:
            # If you are logged in.
            >>> balena.auth.get_actor_id()
        """
        return self.__get_actor_details()["id"]

    def logout(self) -> None:
        """
        This function is used for logging out from balena.

        Examples:
            # If you are logged in.
            >>> balena.auth.logout()
        """
        self._actor_details_cache = None
        self._user_actor_id_cache = None
        self.__settings.remove(TOKEN_KEY)

    def register(self, **creentials: Unpack[CredentialsType]) -> str:
        """
        This function is used for registering to balena.

        Args:
            **credentials: credentials keyword arguments.
                email (str): email to register.
                password (str): Password.

        Returns:
            str: Auth Token for new account.

        Examples:
            >>> credentials = {'email': '<your email>', 'password': '<your password>'}
            >>> balena.auth.register(**credentials)
        """
        return request(
            method="POST",
            settings=self.__settings,
            path="/user/register",
            body=creentials,
            send_token=False,
        )



from typing import List, Union

from .. import exceptions
from ..pine import PineClient
from ..settings import Settings
from ..types import AnyObject
from ..types.models import DeviceTypeType
from ..utils import merge


class DeviceType:
    """
    This class implements user API key model for balena python SDK.

    """

    def __init__(self, pine: PineClient, settings: Settings):
        self.__pine = pine
        self.__settings = settings

    def get(self, id_or_slug: Union[str, int], options: AnyObject = {}) -> DeviceTypeType:
        """
        Get a single device type.

        Args:
            id_or_slug (Union[str, int]): device type slug or alias (string) or id (int).
            options (AnyObject): extra pine options to use.

        Returns:
            DeviceTypeType: Returns the device type
        """

        if id_or_slug is None:
            raise exceptions.InvalidDeviceType(id_or_slug)

        if isinstance(id_or_slug, str):
            device_types = self.get_all(
                merge(
                    {
                        "$top": 1,
                        "$filter": {
                            "device_type_alias": {
                                "$any": {
                                    "$alias": "dta",
                                    "$expr": {
                                        "dta": {
                                            "is_referenced_by__alias": id_or_slug,
                                        },
                                    },
                                },
                            },
                        },
                    },
                    options,
                )
            )
            device_type = None
            if len(device_types) > 0:
                device_type = device_types[0]
        else:
            device_type = self.__pine.get(
                {
                    "resource": "device_type",
                    "id": id_or_slug,
                    "options": options,
                }
            )

        if device_type is None:
            raise exceptions.InvalidDeviceType(id_or_slug)

        return device_type

    def get_all(self, options: AnyObject = {}) -> List[DeviceTypeType]:
        """
        Get all device types.

        Args:
            options (AnyObject): extra pine options to use.

        Returns:
            List[DeviceTypeType]: list contains info of device types.
        """
        opts = merge({"$orderby": "name asc"}, options)
        return self.__pine.get(
            {
                "resource": "device_type",
                "options": opts,
            }
        )

    def get_all_supported(self, options: AnyObject = {}):
        """
        Get all supported device types.

        Args:
            options (AnyObject): extra pine options to use.

        Returns:
            List[DeviceTypeType]: list contains info of all supported device types.
        """

        return self.get_all(
            merge(
                {
                    "$filter": {
                        "is_default_for__application": {
                            "$any": {
                                "$alias": "idfa",
                                "$expr": {
                                    "idfa": {
                                        "is_host": True,
                                        "is_archived": False,
                                        "owns__release": {
                                            "$any": {
                                                "$alias": "r",
                                                "$expr": {
                                                    "r": {
                                                        "status": "success",
                                                        "is_final": True,
                                                        "is_invalidated": False,
                                                    }
                                                },
                                            }
                                        },
                                    }
                                },
                            }
                        }
                    }
                },
                options,
            )
        )

    def get_by_slug_or_name(self, slug_or_name: str, options: AnyObject = {}) -> DeviceTypeType:
        """
        Get a single device type by slug or name.

        Args:
            slug_or_name (str): device type slug or name.
            options (AnyObject): extra pine options to use.

        Returns:
            DeviceTypeType: Returns the device type
        """

        device_types = self.get_all(
            merge(
                {
                    "$top": 1,
                    "$filter": {"$or": [{"name": slug_or_name}, {"slug": slug_or_name}]},
                },
                options,
            )
        )

        device_type = device_types[0] if len(device_types) > 0 else None

        if device_type is None:
            raise exceptions.InvalidDeviceType(slug_or_name)

        return device_type

    def get_name(self, slug: str) -> str:
        """
        Get display name for a device.

        Args:
            slug (str): device type slug.

        """

        return self.get_by_slug_or_name(slug, {"$select": "name"})["name"]

    def get_slug_by_name(self, name: str) -> str:
        """
        Get device slug.

        Args:
            name (str): device type name.

        """

        return self.get_by_slug_or_name(name, {"$select": "slug"})["slug"]



from typing import List, TypedDict, Optional

from ..balena_auth import request
from ..settings import Settings


class GaConfig(TypedDict):
    site: str
    id: str


class ConfigType(TypedDict):
    deployment: Optional[str]
    deviceUrlsBase: str
    adminUrl: str
    gitServerUrl: str
    ga: Optional[GaConfig]
    mixpanelToken: Optional[str]
    intercomAppId: Optional[str]
    recurlyPublicKey: Optional[str]
    DEVICE_ONLINE_ICON: str
    DEVICE_OFFLINE_ICON: str
    signupCodeRequired: bool
    supportedSocialProviders: List[str]


class Config:
    """
    This class implements configuration model for balena python SDK.

    """

    def __init__(self, settings: Settings):
        self.__settings = settings

    def get_all(self) -> ConfigType:
        """
        Get all configuration.

        Returns:
            ConfigType: configuration information.

        Examples:
            >>> balena.models.config.get_all()
        """

        return request(method="GET", path="/config", settings=self.__settings)
